# SPDX-License-Identifier: LGPL-2.1-or-later

# -*- coding: utf8 -*-
# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2018 George Shuklin (amarao)                            *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to return the SVG representation of various shapes."""
## @package svg
# \ingroup draftfunctions
# \brief Provides functions to return the SVG representation of shapes.

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils
import WorkingPlane
from draftfunctions import svgtext
from draftfunctions.svgshapes import get_proj, get_circle, get_path
from draftobjects import layer
from draftutils import params
from draftutils import utils
from draftutils.messages import _wrn, _err

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")
DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")


## \addtogroup draftfunctions
# @{


def get_line_style(line_style, scale):
    """Return a linestyle scaled by a factor."""
    style = None

    if line_style == "Dashed":
        style = params.get_param("svgDashedLine")
    elif line_style == "Dashdot":
        style = params.get_param("svgDashdotLine")
    elif line_style == "Dotted":
        style = params.get_param("svgDottedLine")
    elif line_style:
        if "," in line_style:
            style = line_style

    if style:
        style = style.split(",")
        try:
            # scale dashes
            style = ",".join([str(float(d) / scale) for d in style])
            # print("lstyle ", style)
        except Exception:
            # TODO: trap only specific exception; what is the problem?
            # Bad string specification?
            return "none"
        else:
            return style

    return "none"


def getLineStyle(linestyle, scale):
    """Return a Line style. DEPRECATED. Use get_line_style."""
    utils.use_instead("get_line_style")
    return get_line_style(linestyle, scale)


def get_pattern(pat):
    """Get an SVG pattern."""
    patterns = utils.svg_patterns()

    if pat in patterns:
        return patterns[pat][0]
    return ""


def getPattern(pat):
    """Get an SVG pattern. DEPRECATED."""
    utils.use_instead("get_pattern")
    return get_pattern(pat)


def get_arrow(obj, arrowtype, point, arrowsize, color, linewidth, angle=0):
    """Get the SVG representation from an arrow."""
    svg = ""
    vobj = _get_view_object(obj)

    if not App.GuiUp or vobj is None:
        return svg

    _cx_cy_r = 'cx="{}" cy="{}" r="{}"'.format(point.x, point.y, arrowsize)
    _rotate = "rotate({},{},{})".format(math.degrees(angle), point.x, point.y)
    _transl = "translate({},{})".format(point.x, point.y)
    _scale = "scale({size},{size})".format(size=arrowsize)
    _style = 'style="stroke-miterlimit:4;stroke-dasharray:none;stroke-linecap:square"'

    if arrowtype == "Circle":
        svg += "<circle "
        svg += _cx_cy_r + " "
        svg += 'fill="{}" stroke="{}" '.format("none", color)
        svg += 'style="stroke-width:{};'.format(linewidth)
        svg += 'stroke-miterlimit:4;stroke-dasharray:none;stroke-linecap:square" '
        svg += 'freecad:skip="1"'
        svg += "/>\n"
    elif arrowtype == "Dot":
        svg += "<circle "
        svg += _cx_cy_r + " "
        svg += 'fill="{}" stroke="{}" '.format(color, "none")
        svg += _style + " "
        svg += 'freecad:skip="1"'
        svg += "/>\n"
    elif arrowtype == "Arrow":
        svg += "<path "
        svg += 'transform="'
        svg += _rotate + " "
        svg += _transl + " "
        svg += _scale + '" '
        svg += 'freecad:skip="1" '
        svg += 'fill="{}" stroke="{}" '.format(color, "none")
        svg += _style + " "
        svg += 'd="M 0 0 L 4 1 L 4 -1 Z"'
        svg += "/>\n"
    elif arrowtype == "Tick":
        svg += "<path "
        svg += 'transform="'
        svg += _rotate + " "
        svg += _transl + " "
        svg += _scale + '" '
        svg += 'freecad:skip="1" '
        svg += 'fill="{}" stroke="{}" '.format(color, "none")
        svg += _style + " "
        svg += 'd="M -1 -2 L 0 2 L 1 2 L 0 -2 Z"'
        svg += "/>\n"
    elif arrowtype == "Tick-2":
        svg += "<line "
        svg += 'transform="'
        svg += "rotate({},{},{}) ".format(math.degrees(angle) + 45, point.x, point.y)
        svg += _transl + '" '
        svg += 'freecad:skip="1" '
        svg += 'fill="{}" stroke="{}" '.format("none", color)
        svg += 'style="stroke-dasharray:none;stroke-linecap:square;'
        svg += 'stroke-width:{}" '.format(linewidth)
        svg += 'x1="-{}" y1="0" '.format(2 * arrowsize)
        svg += 'x2="{}" y2="0"'.format(2 * arrowsize)
        svg += "/>\n"
    elif arrowtype == "None":
        svg += ""
    else:
        _wrn("getSVG: arrow type not implemented")

    return svg


def getArrow(obj, arrowtype, point, arrowsize, color, linewidth, angle=0):
    """Get the SVG representation from an arrow. DEPRECATED."""
    utils.use_instead("get_arrow")
    return get_arrow(obj, arrowtype, point, arrowsize, color, linewidth, angle)


def get_overshoot(point, shootsize, color, linewidth, angle=0):
    """Get the SVG representation of a dimension line overshoot."""
    svg = "<line "
    svg += 'transform="'
    svg += "rotate({},{},{}) ".format(math.degrees(angle), point.x, point.y)
    svg += 'translate({},{})" '.format(point.x, point.y)
    svg += 'freecad:skip="1" '
    svg += 'fill="{}" stroke="{}" '.format("none", color)
    svg += 'style="stroke-dasharray:none;stroke-linecap:square;'
    svg += 'stroke-width:{}" '.format(linewidth)
    svg += 'x1="0" y1="0" '
    svg += 'x2="{}" y2="0"'.format(-1 * shootsize)
    svg += "/>\n"
    return svg


def getOvershoot(point, shootsize, color, linewidth, angle=0):
    """Get the SVG representation of a dimension line overshoot. DEPRECATED."""
    utils.use_instead("get_overshoot")
    return get_overshoot(point, shootsize, color, linewidth, angle)


def format_point(coords, action="L"):
    """Return a string with a formatted point."""
    return "{action}{x},{y}".format(x=coords.x, y=coords.y, action=action)


def _svg_shape(svg, obj, plane, fillstyle, pathdata, stroke, linewidth, lstyle):
    """Return the SVG representation of a Part.Shape."""
    if "#" in fillstyle:
        fill = fillstyle
    elif fillstyle == "shape color":
        fill = "#888888"
    elif fillstyle in ("none", None):
        fill = "none"
    else:
        fill = "url(#" + fillstyle + ")"

    svg += get_path(
        obj,
        plane,
        fill,
        pathdata,
        stroke,
        linewidth,
        lstyle,
        fill_opacity=None,
        edges=obj.Edges,
        pathname="",
    )
    return svg


def _svg_dimension(
    obj, plane, scale, linewidth, fontsize, stroke, tstroke, pointratio, techdraw, rotation
):
    """Return the SVG representation of a linear dimension."""
    if not App.GuiUp:
        _wrn("'{}': SVG can only be generated " "in GUI mode".format(obj.Label))
        return ""

    vobj = _get_view_object(obj)

    if not hasattr(vobj, "Proxy") or not vobj.Proxy:
        _err("'{}': doesn't have Proxy, " "SVG cannot be generated".format(obj.Label))
        return ""

    prx = vobj.Proxy

    if not hasattr(prx, "p1"):
        _err("'{}': doesn't have points, " "SVG cannot be generated".format(obj.Label))
        return ""

    ts = len(prx.string) * vobj.FontSize.Value / 4.0
    rm = (prx.p3 - prx.p2).Length / 2.0 - ts

    _diff32 = prx.p3 - prx.p2
    _diff23 = prx.p2 - prx.p3

    _v32 = DraftVecUtils.scaleTo(_diff32, rm)
    _v23 = DraftVecUtils.scaleTo(_diff23, rm)

    p2a = get_proj(prx.p2 + _v32, plane)
    p2b = get_proj(prx.p3 + _v23, plane)
    p1 = get_proj(prx.p1, plane)
    p2 = get_proj(prx.p2, plane)
    p3 = get_proj(prx.p3, plane)
    p4 = get_proj(prx.p4, plane)

    tbase = get_proj(prx.tbase, plane)
    r = prx.textpos.rotation.getValue().getValue()
    _rv = App.Rotation(r[0], r[1], r[2], r[3])
    rv = _rv.multVec(App.Vector(1, 0, 0))
    angle = -DraftVecUtils.angle(get_proj(rv, plane))
    # angle = -DraftVecUtils.angle(p3.sub(p2))

    svg = ""
    nolines = False
    if hasattr(vobj, "ShowLine"):
        if not vobj.ShowLine:
            nolines = True

    # drawing lines
    if not nolines:
        svg += "<path "

    if vobj.DisplayMode == "World":
        tangle = angle
        if tangle > math.pi / 2:
            tangle = tangle - math.pi
        # elif (tangle <= -math.pi/2) or (tangle > math.pi/2):
        #    tangle = tangle + math.pi

        if rotation != 0:
            # print("dim: tangle:", tangle,
            #       " rot: ", rotation,
            #       " text: ", prx.string)
            if abs(tangle + math.radians(rotation)) < 0.0001:
                tangle += math.pi
                _v = App.Vector(0, 2.0 / scale, 0)
                _rot = DraftVecUtils.rotate(_v, tangle)
                tbase = tbase + _rot

        if not nolines:
            svg += 'd="M ' + str(p1.x) + " " + str(p1.y) + " "
            svg += "L " + str(p2.x) + " " + str(p2.y) + " "
            svg += "L " + str(p3.x) + " " + str(p3.y) + " "
            svg += "L " + str(p4.x) + " " + str(p4.y) + '" '
    else:
        tangle = 0
        if rotation != 0:
            tangle = -math.radians(rotation)

        tbase = tbase + App.Vector(0, -2.0 / scale, 0)
        if not nolines:
            svg += 'd="M ' + str(p1.x) + " " + str(p1.y) + " "
            svg += "L " + str(p2.x) + " " + str(p2.y) + " "
            svg += "L " + str(p2a.x) + " " + str(p2a.y) + " "
            svg += "M " + str(p2b.x) + " " + str(p2b.y) + " "
            svg += "L " + str(p3.x) + " " + str(p3.y) + " "
            svg += "L " + str(p4.x) + " " + str(p4.y) + '" '

    if not nolines:
        svg += 'fill="none" stroke="'
        svg += stroke + '" '
        svg += 'stroke-width="' + str(linewidth) + ' px" '
        svg += 'style="stroke-width:' + str(linewidth)
        svg += ';stroke-miterlimit:4;stroke-dasharray:none;stroke-linecap:square" '
        svg += 'freecad:basepoint1="' + str(p1.x) + " " + str(p1.y) + '" '
        svg += 'freecad:basepoint2="' + str(p4.x) + " " + str(p4.y) + '" '
        svg += 'freecad:dimpoint="' + str(p2.x) + " " + str(p2.y) + '"'
        svg += "/>\n"

        # drawing dimension and extension lines overshoots
        if hasattr(vobj, "DimOvershoot") and vobj.DimOvershoot.Value:
            shootsize = vobj.DimOvershoot.Value / pointratio
            svg += get_overshoot(p2, shootsize, stroke, linewidth, angle)
            svg += get_overshoot(p3, shootsize, stroke, linewidth, angle + math.pi)

        if hasattr(vobj, "ExtOvershoot") and vobj.ExtOvershoot.Value:
            shootsize = vobj.ExtOvershoot.Value / pointratio
            shootangle = -DraftVecUtils.angle(p1 - p2)
            svg += get_overshoot(p2, shootsize, stroke, linewidth, shootangle)
            svg += get_overshoot(p3, shootsize, stroke, linewidth, shootangle)

        # drawing arrows
        if (
            hasattr(vobj, "ArrowTypeStart")
            and hasattr(vobj, "ArrowTypeEnd")
            and hasattr(vobj, "ArrowSizeStart")
            and hasattr(vobj, "ArrowSizeEnd")
        ):
            if getattr(vobj, "FlipArrows", False):
                angle = angle + math.pi
            if getattr(vobj, "FlipText", False):
                angle = angle + math.pi

            svg += get_arrow(
                obj,
                vobj.ArrowTypeStart,
                p2,
                vobj.ArrowSizeStart.Value / pointratio,
                stroke,
                linewidth,
                angle,
            )

            svg += get_arrow(
                obj,
                vobj.ArrowTypeEnd,
                p3,
                vobj.ArrowSizeEnd.Value / pointratio,
                stroke,
                linewidth,
                angle + math.pi,
            )

    # drawing text
    svg += svgtext.get_text(
        plane, techdraw, tstroke, fontsize, vobj.FontName, tangle, tbase, prx.string
    )

    return svg


def get_svg(
    obj,
    scale=1,
    linewidth=0.35,
    fontsize=12,
    fillstyle="shape color",
    direction=None,
    linestyle=None,
    color=None,
    linespacing=None,
    techdraw=False,
    rotation=0,
    fillspaces=False,
    override=True,
):
    """Return a string containing an SVG representation of the object.

    Paramaeters
    -----------
    scale: float, optional
        It defaults to 1.
        It allows scaling line widths down, so they are resolution-independent.

    linewidth: float, optional
        It defaults to 0.35.

    fontsize: float, optional
        It defaults to 12, which is interpreted as `pt` unit (points).
        It is used if the given object contains any text.

    fillstyle: str, optional
        It defaults to 'shape color'.

    direction: Base::Vector3, optional
        It defaults to `None`.
        It is an arbitrary projection vector or a `WorkingPlane.PlaneBase`
        instance.

    linestyle: optional
        It defaults to `None`.

    color: optional
        It defaults to `None`.

    linespacing: float, optional
        It defaults to `None`.

    techdraw: bool, optional
        It defaults to `False`.
        If it is `True`, it sets some options for generating SVG strings
        for displaying inside TechDraw.

    rotation: float, optional
        It defaults to 0.

    fillspaces: bool, optional
        It defaults to `False`.

    override: bool, optional
        It defaults to `True`.
    """
    # If this is a group, recursively call this function to gather
    # all the SVG strings from the contents of the group.
    if (
        obj.isDerivedFrom("App::DocumentObjectGroup")
        or utils.get_type(obj) in ["Layer", "BuildingPart", "IfcGroup"]
        or obj.isDerivedFrom("App::LinkGroup")
        or (
            obj.isDerivedFrom("App::Link")
            and obj.LinkedObject.isDerivedFrom("App::DocumentObjectGroup")
        )
    ):

        hidden_doc = None

        if obj.isDerivedFrom("App::LinkGroup") or (
            obj.isDerivedFrom("App::Link")
            and obj.LinkedObject.isDerivedFrom("App::DocumentObjectGroup")
        ):
            if obj.Placement.isIdentity():
                if obj.isDerivedFrom("App::LinkGroup"):
                    group = obj.ElementList
                else:
                    group = obj.Group
            else:
                # Hidden doc hack:
                hidden_doc = App.newDocument(name="hidden", hidden=True, temp=True)
                new = hidden_doc.copyObject(obj, True)
                pla = new.Placement
                new.Placement = App.Placement()
                if new.isDerivedFrom("App::LinkGroup"):
                    group = new.ElementList
                else:
                    group = new.Group
                for child in group:
                    child.Placement = pla * child.Placement
        else:
            group = obj.Group

        svg = ""
        for child in group:
            svg += get_svg(
                child,
                scale,
                linewidth,
                fontsize,
                fillstyle,
                direction,
                linestyle,
                color,
                linespacing,
                techdraw,
                rotation,
                fillspaces,
                override,
            )

        if hidden_doc is not None:
            try:
                App.closeDocument(hidden_doc.Name)
            except:
                pass

        return svg

    # Handle Links to texts and dimensions. These Links do not have a Shape.
    if obj.isDerivedFrom("App::Link") and obj.LinkedObject and not hasattr(obj, "Shape"):
        # Hidden doc hack:
        hidden_doc = App.newDocument(name="hidden", hidden=True, temp=True)
        new = hidden_doc.copyObject(obj.LinkedObject, True)
        if utils.get_type(new) in ("Dimension", "LinearDimension", "AngularDimension"):
            new.Proxy.transform(new, obj.Placement)
        elif utils.get_type(new) == "Text" or obj.LinkTransform:
            new.Placement = obj.Placement * new.Placement
        else:
            new.Placement = obj.Placement
        svg = get_svg(
            new,
            scale,
            linewidth,
            fontsize,
            fillstyle,
            direction,
            linestyle,
            color,
            linespacing,
            techdraw,
            rotation,
            fillspaces,
            override,
        )
        try:
            App.closeDocument(hidden_doc.Name)
        except:
            pass
        return svg

    vobj = _get_view_object(obj)

    pathdata = []
    svg = ""
    linewidth = float(linewidth) / scale
    if not override:
        if vobj is not None and hasattr(vobj, "LineWidth"):
            if hasattr(vobj.LineWidth, "Value"):
                lw = vobj.LineWidth.Value
            else:
                lw = vobj.LineWidth
            linewidth = lw * linewidth

    fontsize = (float(fontsize) / scale) / 2
    if linespacing:
        linespacing = float(linespacing) / scale
    else:
        linespacing = 0.5

    # print(obj.Label, "line spacing", linespacing, "scale", scale)

    # The number of times the dots are smaller than the arrow size
    pointratio = 0.75
    plane = None

    if direction:
        if isinstance(direction, App.Vector):
            if direction != App.Vector(0, 0, 0):
                plane = WorkingPlane.PlaneBase()
                plane.align_to_point_and_axis_svg(
                    App.Vector(0, 0, 0), direction.negative().negative(), 0
                )
            else:
                raise ValueError("'direction' cannot be: Vector(0, 0, 0)")
        elif isinstance(direction, WorkingPlane.PlaneBase):
            plane = direction

    stroke = "#000000"
    tstroke = stroke
    if color and override:
        if "#" in color:
            stroke = color
        else:
            stroke = utils.get_rgb(color)
        tstroke = stroke
    elif App.GuiUp:
        # find print color
        pc = get_print_color(obj)
        if pc:
            stroke = utils.get_rgb(pc)
        # get line color
        elif vobj is not None:
            if hasattr(vobj, "LineColor"):
                stroke = utils.get_rgb(vobj.LineColor)
            elif hasattr(vobj, "TextColor"):
                stroke = utils.get_rgb(vobj.TextColor)
            if hasattr(vobj, "TextColor"):
                tstroke = utils.get_rgb(vobj.TextColor)

    lstyle = "none"
    if override:
        lstyle = get_line_style(linestyle, scale)
    else:
        if vobj is not None and hasattr(vobj, "DrawStyle"):
            lstyle = get_line_style(vobj.DrawStyle, scale)

    if not obj:
        pass

    elif isinstance(obj, Part.Shape):
        svg = _svg_shape(svg, obj, plane, fillstyle, pathdata, stroke, linewidth, lstyle)

    elif utils.get_type(obj) in ["Dimension", "LinearDimension"] or (
        utils.get_type(obj) == "IfcAnnotation" and obj.ObjectType == "DIMENSION"
    ):
        svg = _svg_dimension(
            obj, plane, scale, linewidth, fontsize, stroke, tstroke, pointratio, techdraw, rotation
        )

    elif utils.get_type(obj) == "AngularDimension":
        if not App.GuiUp:
            _wrn("Export of dimensions to SVG is only available in GUI mode")

        if App.GuiUp:
            if vobj.Proxy:
                if hasattr(vobj.Proxy, "circle"):
                    prx = vobj.Proxy

                    # drawing arc
                    fill = "none"
                    if vobj.DisplayMode == "World":
                        svg += get_path(
                            obj,
                            plane,
                            fill,
                            pathdata,
                            stroke,
                            linewidth,
                            lstyle,
                            fill_opacity=None,
                            edges=[prx.circle],
                        )
                    else:
                        if hasattr(prx, "circle1"):
                            svg += get_path(
                                obj,
                                plane,
                                fill,
                                pathdata,
                                stroke,
                                linewidth,
                                lstyle,
                                fill_opacity=None,
                                edges=[prx.circle1],
                            )
                            svg += get_path(
                                obj,
                                plane,
                                fill,
                                pathdata,
                                stroke,
                                linewidth,
                                lstyle,
                                fill_opacity=None,
                                edges=[prx.circle2],
                            )
                        else:
                            svg += get_path(
                                obj,
                                plane,
                                fill,
                                pathdata,
                                stroke,
                                linewidth,
                                lstyle,
                                fill_opacity=None,
                                edges=[prx.circle],
                            )

                    # drawing arrows
                    if (
                        hasattr(vobj, "ArrowTypeStart")
                        and hasattr(vobj, "ArrowTypeEnd")
                        and hasattr(vobj, "ArrowSizeStart")
                        and hasattr(vobj, "ArrowSizeEnd")
                    ):
                        p2 = get_proj(prx.p2, plane)
                        p3 = get_proj(prx.p3, plane)
                        arrowsizestart = vobj.ArrowSizeStart.Value / pointratio
                        halfstartarrowlength = 2 * arrowsizestart
                        startarrowangle = 2 * math.asin(
                            halfstartarrowlength / prx.circle.Curve.Radius
                        )
                        arrowsizeend = vobj.ArrowSizeEnd.Value / pointratio
                        halfendarrowlength = 2 * arrowsizeend
                        endarrowangle = 2 * math.asin(halfendarrowlength / prx.circle.Curve.Radius)
                        if hasattr(vobj, "FlipArrows") and vobj.FlipArrows:
                            startarrowangle = -startarrowangle
                            endarrowangle = -endarrowangle

                        _v1a = prx.circle.valueAt(prx.circle.FirstParameter + startarrowangle)
                        _v1b = prx.circle.valueAt(prx.circle.FirstParameter)

                        _v2a = prx.circle.valueAt(prx.circle.LastParameter - endarrowangle)
                        _v2b = prx.circle.valueAt(prx.circle.LastParameter)

                        u1 = get_proj(_v1a - _v1b, plane)
                        u2 = get_proj(_v2a - _v2b, plane)
                        angle1 = -DraftVecUtils.angle(u1)
                        angle2 = -DraftVecUtils.angle(u2)

                        svg += get_arrow(
                            obj, vobj.ArrowTypeStart, p2, arrowsizestart, stroke, linewidth, angle1
                        )
                        svg += get_arrow(
                            obj, vobj.ArrowTypeEnd, p3, arrowsizeend, stroke, linewidth, angle2
                        )

                    # drawing text
                    if vobj.DisplayMode == "World":
                        _diff = prx.circle.LastParameter - prx.circle.FirstParameter
                        t = prx.circle.tangentAt(prx.circle.FirstParameter + _diff / 2.0)
                        t = get_proj(t, plane)
                        tangle = -DraftVecUtils.angle(t)
                        if (tangle <= -math.pi / 2) or (tangle > math.pi / 2):
                            tangle = tangle + math.pi

                        _diff = prx.circle.LastParameter - prx.circle.FirstParameter
                        _va = prx.circle.valueAt(prx.circle.FirstParameter + _diff / 2.0)
                        tbase = get_proj(_va, plane)

                        _v = App.Vector(0, 2.0 / scale, 0)
                        tbase = tbase + DraftVecUtils.rotate(_v, tangle)
                        # print(tbase)
                    else:
                        tangle = 0
                        tbase = get_proj(prx.tbase, plane)

                    svg += svgtext.get_text(
                        plane, techdraw, tstroke, fontsize, vobj.FontName, tangle, tbase, prx.string
                    )

    elif utils.get_type(obj) == "Label":

        if not App.GuiUp:
            _wrn("Export of texts to SVG is only available in GUI mode")

        if App.GuiUp:

            if getattr(vobj, "Line", True):
                # Some Labels may have no Line property
                # Draw multisegment line
                proj_points = list(map(lambda x: get_proj(x, plane), obj.Points))
                path_dir_list = [format_point(proj_points[0], action="M")]
                path_dir_list += map(format_point, proj_points[1:])
                path_dir_str = " ".join(path_dir_list)

                svg_path = "<path "
                svg_path += 'fill="none" '
                svg_path += 'stroke="{}" '.format(stroke)
                svg_path += 'stroke-width="{}" '.format(linewidth)
                svg_path += 'stroke-linecap="square" '
                svg_path += 'd="{}"'.format(path_dir_str)
                svg_path += "/>"
                svg += svg_path

                # Draw arrow.
                # We are different here from 3D view
                # if Line is set to 'off', no arrow is drawn
                if hasattr(vobj, "ArrowTypeStart") and (
                    hasattr(vobj, "ArrowSizeStart") and len(obj.Points) > 1
                ):
                    last_segment = App.Vector(obj.Points[-1] - obj.Points[-2])
                    _v = get_proj(last_segment, plane)
                    angle = -DraftVecUtils.angle(_v) + math.pi
                    svg += get_arrow(
                        obj,
                        vobj.ArrowTypeStart,
                        proj_points[-1],
                        vobj.ArrowSizeStart.Value / pointratio,
                        stroke,
                        linewidth,
                        angle,
                    )

            fontname = vobj.FontName
            position = get_proj(obj.Placement.Base, plane)
            rotation = obj.Placement.Rotation
            justification = vobj.Justification
            text = obj.Text
            svg += svgtext.get_text(
                plane,
                techdraw,
                tstroke,
                fontsize,
                fontname,
                rotation,
                position,
                text,
                linespacing,
                justification,
            )

    elif utils.get_type(obj) in ["Annotation", "DraftText", "Text"] or (
        utils.get_type(obj) == "IfcAnnotation" and obj.ObjectType == "TEXT"
    ):
        # returns an svg representation of a document annotation
        if not App.GuiUp:
            _wrn("Export of texts to SVG is only available in GUI mode")

        if App.GuiUp:
            n = vobj.FontName
            if utils.get_type(obj) == "Annotation":
                p = get_proj(obj.Position, plane)
                r = vobj.Rotation.getValueAs("rad")
                t = obj.LabelText
            else:  # DraftText (old) or Text (new, 0.19)
                p = get_proj(obj.Placement.Base, plane)
                r = obj.Placement.Rotation
                t = obj.Text

            j = vobj.Justification
            svg += svgtext.get_text(plane, techdraw, tstroke, fontsize, n, r, p, t, linespacing, j)

    elif utils.get_type(obj) == "Axis":
        # returns the SVG representation of an Arch Axis system
        if not App.GuiUp:
            _wrn("Export of axes to SVG is only available in GUI mode")

        if App.GuiUp:
            fn = vobj.FontName
            fill = "none"
            rad = vobj.BubbleSize.Value / 2
            n = 0
            for e in obj.Shape.Edges:
                svg += get_path(
                    obj,
                    plane,
                    fill,
                    pathdata,
                    stroke,
                    linewidth,
                    lstyle,
                    fill_opacity=None,
                    edges=[e],
                )
            for t in vobj.Proxy.getTextData():
                pos = t[1].add(App.Vector(0, -fontsize / 2, 0))
                svg += svgtext.get_text(
                    plane, techdraw, tstroke, fontsize, fn, 0.0, pos, t[0], 1.0, "center"
                )
            for b in vobj.Proxy.getShapeData():
                if hasattr(b, "Curve") and isinstance(b.Curve, Part.Circle):
                    svg += get_circle(plane, fill, stroke, linewidth, "none", b)
                else:
                    sfill = stroke
                    svg += get_path(
                        obj,
                        plane,
                        sfill,
                        pathdata,
                        stroke,
                        linewidth,
                        "none",
                        fill_opacity=None,
                        edges=b.Edges,
                    )

    elif utils.get_type(obj) == "Pipe":
        fill = stroke
        if obj.Base and obj.Diameter:
            svg += get_path(
                obj,
                plane,
                fill,
                pathdata,
                stroke,
                linewidth,
                lstyle,
                fill_opacity=None,
                edges=obj.Base.Shape.Edges,
            )
        for f in obj.Shape.Faces:
            if len(f.Edges) == 1:
                if isinstance(f.Edges[0].Curve, Part.Circle):
                    svg += get_circle(plane, fill, stroke, linewidth, lstyle, f.Edges[0])

    elif utils.get_type(obj) == "Rebar":
        fill = "none"
        basewire = obj.Base.Shape.Wires[0].copy()
        # Not applying rounding because the results are not correct
        # if hasattr(obj, "Rounding") and obj.Rounding:
        #     basewire = DraftGeomUtils.filletWire(
        #         basewire, obj.Rounding * obj.Diameter.Value
        #     )
        wires = []
        for placement in obj.PlacementList:
            wire = basewire.copy()
            wire.Placement = placement.multiply(basewire.Placement)
            wires.append(wire)
        svg += get_path(
            obj, plane, fill, pathdata, stroke, linewidth, lstyle, fill_opacity=None, wires=wires
        )

    elif utils.get_type(obj) == "PipeConnector":
        pass

    elif utils.get_type(obj) == "Space":
        fill_opacity = 1

        # returns an SVG fragment for the text of a space
        if not App.GuiUp:
            _wrn("Export of spaces to SVG is only available in GUI mode")

        if App.GuiUp:
            if fillspaces and hasattr(obj, "Proxy"):
                if not hasattr(obj.Proxy, "face"):
                    obj.Proxy.getArea(obj, notouch=True)
                if hasattr(obj.Proxy, "face"):
                    # setting fill
                    if App.GuiUp and hasattr(vobj, "ShapeColor"):
                        fill = utils.get_rgb(vobj.ShapeColor, testbw=False)
                        fill_opacity = 1 - vobj.Transparency / 100.0
                    else:
                        fill = "#888888"
                    svg += get_path(
                        obj,
                        plane,
                        fill,
                        pathdata,
                        stroke,
                        linewidth,
                        lstyle,
                        fill_opacity=fill_opacity,
                        wires=[obj.Proxy.face.OuterWire],
                    )
            c = utils.get_rgb(vobj.TextColor)
            n = vobj.FontName
            a = 0
            if rotation != 0:
                a = math.radians(rotation)

            t1 = vobj.Proxy.text1.string.getValues()
            t2 = vobj.Proxy.text2.string.getValues()
            scale = vobj.FirstLine.Value / vobj.FontSize.Value
            f1 = fontsize * scale

            if round(plane.axis.getAngle(App.Vector(0, 0, 1)), 2) not in [0, 3.14]:
                # if not in XY view, place the label at center
                p2 = obj.Shape.CenterOfMass
            else:
                _v = vobj.Proxy.coords.translation.getValue().getValue()
                p2 = obj.Placement.multVec(App.Vector(_v))

            _h = vobj.Proxy.header.translation.getValue().getValue()
            lspc = App.Vector(_h)
            p1 = p2 + lspc
            j = vobj.TextAlign
            t3 = svgtext.get_text(
                plane, techdraw, c, f1, n, a, get_proj(p1, plane), t1, linespacing, j, flip=True
            )
            svg += t3
            if t2:
                ofs = App.Vector(0, -lspc.Length, 0)
                if a:
                    Z = App.Vector(0, 0, 1)
                    ofs = App.Rotation(Z, -rotation).multVec(ofs)
                t4 = svgtext.get_text(
                    plane,
                    techdraw,
                    c,
                    fontsize,
                    n,
                    a,
                    get_proj(p1, plane).add(ofs),
                    t2,
                    linespacing,
                    j,
                    flip=True,
                )
                svg += t4

    elif hasattr(obj, "Shape"):
        # In the past we tested for a Part Feature
        # elif obj.isDerivedFrom('Part::Feature'):
        #
        # however, this doesn't work for App::Links; instead we
        # test for a 'Shape'. All Part::Features should have a Shape,
        # and App::Links can have one as well.
        if obj.Shape.isNull():
            return ""

        fill_opacity = 1
        # setting fill
        if obj.Shape.Faces:
            if App.GuiUp:
                try:
                    m = vobj.DisplayMode
                except AttributeError:
                    m = None

                if m != "Wireframe":
                    if (fillstyle == "shape color") and hasattr(vobj, "ShapeColor"):
                        fill = utils.get_rgb(vobj.ShapeColor, testbw=False)
                        fill_opacity = 1 - vobj.Transparency / 100.0
                    elif fillstyle in ("none", None):
                        fill = "none"
                    else:
                        fill = "url(#" + fillstyle + ")"
                        svg += get_pattern(fillstyle)
                else:
                    fill = "none"
            else:
                fill = "#888888"
        else:
            fill = "none"

        if len(obj.Shape.Vertexes) > 1:
            wiredEdges = []
            if obj.Shape.Faces:
                for i, f in enumerate(obj.Shape.Faces):
                    # place outer wire first
                    wires = [f.OuterWire]
                    wires.extend([w for w in f.Wires if w.hashCode() != f.OuterWire.hashCode()])
                    svg += get_path(
                        obj,
                        plane,
                        fill,
                        pathdata,
                        stroke,
                        linewidth,
                        lstyle,
                        fill_opacity=fill_opacity,
                        wires=f.Wires,
                        pathname="%s_f%04d" % (obj.Name, i),
                    )
                    wiredEdges.extend(f.Edges)
            else:
                for i, w in enumerate(obj.Shape.Wires):
                    svg += get_path(
                        obj,
                        plane,
                        fill,
                        pathdata,
                        stroke,
                        linewidth,
                        lstyle,
                        fill_opacity=fill_opacity,
                        edges=w.Edges,
                        pathname="%s_w%04d" % (obj.Name, i),
                    )
                    wiredEdges.extend(w.Edges)

            if len(wiredEdges) != len(obj.Shape.Edges):
                fill = "none"  # Required if obj has a face. Edges processed here have no face.

                def get_edge_descriptor(edge):
                    return (
                        str(edge.Curve),
                        str(edge.Vertexes[0].Point),
                        str(edge.Vertexes[-1].Point),
                    )

                wiredEdgesSet = set([get_edge_descriptor(e) for e in wiredEdges])
                for i, e in enumerate(obj.Shape.Edges):
                    if get_edge_descriptor(e) not in wiredEdgesSet:
                        svg += get_path(
                            obj,
                            plane,
                            fill,
                            pathdata,
                            stroke,
                            linewidth,
                            lstyle,
                            fill_opacity=fill_opacity,
                            edges=[e],
                            pathname="%s_nwe%04d" % (obj.Name, i),
                        )
        else:
            # closed circle or spline
            if obj.Shape.Edges:
                if isinstance(obj.Shape.Edges[0].Curve, Part.Circle):
                    svg = get_circle(plane, fill, stroke, linewidth, lstyle, obj.Shape.Edges[0])
                else:
                    svg = get_path(
                        obj,
                        plane,
                        fill,
                        pathdata,
                        stroke,
                        linewidth,
                        lstyle,
                        fill_opacity=fill_opacity,
                        edges=obj.Shape.Edges,
                    )

        if (
            App.GuiUp
            and hasattr(vobj, "ArrowTypeStart")
            and hasattr(vobj, "ArrowTypeEnd")
            and hasattr(vobj, "ArrowSizeStart")
            and hasattr(vobj, "ArrowSizeEnd")
            and len(obj.Shape.Vertexes) > 1
        ):
            # Draft_Wire
            p1 = get_proj(obj.Shape.Vertexes[0].Point, plane)
            p2 = get_proj(obj.Shape.Vertexes[1].Point, plane)
            svg += get_arrow(
                obj,
                vobj.ArrowTypeStart,
                p1,
                vobj.ArrowSizeStart.Value / pointratio,
                stroke,
                linewidth,
                -DraftVecUtils.angle(p2 - p1),
            )
            p1 = get_proj(obj.Shape.Vertexes[-1].Point, plane)
            p2 = get_proj(obj.Shape.Vertexes[-2].Point, plane)
            svg += get_arrow(
                obj,
                vobj.ArrowTypeEnd,
                p1,
                vobj.ArrowSizeEnd.Value / pointratio,
                stroke,
                linewidth,
                -DraftVecUtils.angle(p2 - p1),
            )

    # techdraw expects bottom-to-top coordinates
    if techdraw:
        svg = '<g transform ="scale(1,-1)">\n    ' + svg + "</g>\n"

    return svg


def _get_view_object(obj):
    if (
        obj.isDerivedFrom("App::Link")
        and hasattr(obj, "ViewObject")
        and hasattr(obj.ViewObject, "OverrideMaterial")
        and not obj.ViewObject.OverrideMaterial
        and hasattr(obj.LinkedObject, "ViewObject")
    ):
        return obj.LinkedObject.ViewObject
    if hasattr(obj, "ViewObject"):
        return obj.ViewObject
    return None


def get_print_color(obj):
    """Return the print color of the parent layer, if available."""
    # Layers are not in the Inlist of obj because a layer's Group is App::PropertyLinkListHidden:
    lyr = layer.get_layer(obj)
    if lyr is None:
        return None
    if lyr.ViewObject.UsePrintColor:
        return lyr.ViewObject.LinePrintColor
    return None


def getSVG(
    obj,
    scale=1,
    linewidth=0.35,
    fontsize=12,
    fillstyle="shape color",
    direction=None,
    linestyle=None,
    color=None,
    linespacing=None,
    techdraw=False,
    rotation=0,
    fillSpaces=False,
    override=True,
):
    """Return SVG string of the object. DEPRECATED. Use 'get_svg'."""
    utils.use_instead("get_svg")
    return get_svg(
        obj,
        scale=scale,
        linewidth=linewidth,
        fontsize=fontsize,
        fillstyle=fillstyle,
        direction=direction,
        linestyle=linestyle,
        color=color,
        linespacing=linespacing,
        techdraw=techdraw,
        rotation=rotation,
        fillspaces=fillSpaces,
        override=override,
    )


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2012 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"The FreeCAD Arch Vector Rendering Module"

## @package ArchVRM
#  \ingroup ARCH
#  \brief The Arch Vector Rendering Module
#
#  This module provides the Renderer Class, that allows one to
#  produce SVG renderings of projected shapes, with filled faces.
#  It is used by the "Solid" mode of Arch views in TechDraw and Drawing,
#  and is called from ArchSectionPlane code.

import math

import FreeCAD
import ArchCommands
import Draft
import DraftVecUtils
import DraftGeomUtils
import Part

from draftutils import params

MAXLOOP = 10  # the max number of loop before abort

# WARNING: in this module, faces are lists whose first item is the actual OCC face, the
# other items being additional information such as color, etc.

DEBUG = params.get_param_arch("ShowVRMDebug")


class Renderer:
    "A renderer object"

    def __init__(self, wp=None):
        """
        Creates a renderer with a default Draft WorkingPlane
        Use like this:

        import ArchVRM
        p = ArchVRM.Renderer()
        p.add(App.ActiveDocument.ActiveObject)
        p.sort()
        p.buildDummy()
        """

        self.reset()
        if wp:
            self.wp = wp
        else:
            import WorkingPlane

            self.wp = WorkingPlane.PlaneBase()

        if DEBUG:
            print("Renderer initialized on " + str(self.wp))

    def __str__(self):
        return "Arch Renderer: " + str(len(self.faces)) + " faces projected on " + str(self.wp)

    def reset(self):
        "removes all faces from this renderer"
        self.objects = []
        self.shapes = []
        self.faces = []
        self.resetFlags()

    def resetFlags(self):
        "resets all flags of this renderer"
        self.oriented = False
        self.trimmed = False
        self.sorted = False
        self.iscut = False
        self.joined = False
        self.sections = []
        self.hiddenEdges = []

    def setWorkingPlane(self, wp):
        "sets a Draft WorkingPlane or Placement for this renderer"
        if isinstance(wp, FreeCAD.Placement):
            self.wp.align_to_placement(wp)
        else:
            self.wp = wp
        if DEBUG:
            print("Renderer set on " + str(self.wp))

    def addFaces(self, faces, color=(0.9, 0.9, 0.9, 1.0)):
        "add individual faces to this renderer, optionally with a color"
        if DEBUG:
            print(
                "adding ",
                len(faces),
                " faces. Warning, these will get lost if using cut() or join()",
            )
        for f in faces:
            self.faces.append([f, color])
        self.resetFlags()

    def addObjects(self, objs):
        "add objects to this renderer"
        for o in objs:
            if o.isDerivedFrom("Part::Feature"):
                self.objects.append(o)
                color = o.ViewObject.ShapeColor
                if o.Shape.Faces:
                    self.shapes.append([o.Shape, color])
                    for f in o.Shape.Faces:
                        self.faces.append([f, color])
        self.resetFlags()
        if DEBUG:
            print("adding ", len(self.objects), " objects, ", len(self.faces), " faces")

    def addShapes(self, shapes, color=(0.9, 0.9, 0.9, 1.0)):
        "add shapes to this renderer, optionally with a color. Warning, these will get lost if using join()"
        if DEBUG:
            print("adding ", len(shapes), " shapes")
        for s in shapes:
            if s.Faces:
                self.shapes.append([s, color])
                for f in s.Faces:
                    self.faces.append([f, color])
        self.resetFlags()

    def info(self):
        "Prints info about the contents of this renderer"
        r = str(self) + "\n"
        r += "oriented: " + str(self.oriented) + "\n"
        r += "trimmed: " + str(self.trimmed) + "\n"
        r += "sorted: " + str(self.sorted) + "\n"
        r += "contains " + str(len(self.faces)) + " faces\n"
        for i in range(len(self.faces)):
            r += "  face " + str(i) + " : center " + str(self.faces[i][0].CenterOfMass)
            r += " : normal " + str(self.faces[i][0].normalAt(0, 0))
            r += ", " + str(len(self.faces[i][0].Vertexes)) + " verts"
            r += ", color: " + self.getFill(self.faces[i][1]) + "\n"
        return r

    def addLabels(self):
        "Add labels on the model to identify faces"
        c = 0
        for f in self.faces:
            l = FreeCAD.ActiveDocument.addObject("App::AnnotationLabel", "facelabel")
            l.BasePosition = f[0].CenterOfMass
            l.LabelText = str(c)
            c += 1

    def isVisible(self, face):
        "returns True if the given face points in the view direction"
        normal = face[0].normalAt(0, 0)
        if DEBUG:
            print(
                "checking face normal ",
                normal,
                " against ",
                self.wp.axis,
                " : ",
                math.degrees(normal.getAngle(self.wp.axis)),
            )
        if normal.getAngle(self.wp.axis) < math.pi / 2:
            return True
        return False

    def reorient(self):
        "reorients the faces on the WP"
        # print("VRM: start reorient")
        if not self.faces:
            return
        self.faces = [self.projectFace(f) for f in self.faces]
        if self.sections:
            self.sections = [self.projectFace(f) for f in self.sections]
        if self.hiddenEdges:
            self.hiddenEdges = [self.projectEdge(e) for e in self.hiddenEdges]
        self.oriented = True
        # print("VRM: end reorient")

    def removeHidden(self):
        "removes faces pointing outwards"
        if not self.faces:
            return
        faces = []
        for f in self.faces:
            if self.isVisible(f):
                faces.append(f)
        if DEBUG:
            print(len(self.faces) - len(faces), " faces removed, ", len(faces), " faces retained")
        self.faces = faces
        self.trimmed = True

    def projectFace(self, face):
        "projects a single face on the WP"
        # print("VRM: projectFace start: ",len(face[0].Vertexes)," verts, ",len(face[0].Edges)," edges")
        wires = []
        if not face[0].Wires:
            if DEBUG:
                print("Error: Unable to project face on the WP")
            return None
        norm = face[0].normalAt(0, 0)
        for w in face[0].Wires:
            verts = []
            edges = Part.__sortEdges__(w.Edges)
            # print(len(edges)," edges after sorting")
            for e in edges:
                v = e.Vertexes[0].Point
                # print(v)
                v = self.wp.get_local_coords(v)
                verts.append(v)
            verts.append(verts[0])
            if len(verts) > 2:
                # print("new wire with ",len(verts))
                wires.append(Part.makePolygon(verts))
        try:
            sh = ArchCommands.makeFace(wires)
        except Exception:
            if DEBUG:
                print("Error: Unable to project face on the WP")
            return None
        else:
            # restoring flipped normals
            vnorm = self.wp.get_local_coords(norm)
            if vnorm.getAngle(sh.normalAt(0, 0)) > 1:
                sh.reverse()
            # print("VRM: projectFace end: ",len(sh.Vertexes)," verts")
            return [sh] + face[1:]

    def projectEdge(self, edge):
        "projects a single edge on the WP"
        if len(edge.Vertexes) > 1:
            v1 = self.wp.get_local_coords(edge.Vertexes[0].Point)
            v2 = self.wp.get_local_coords(edge.Vertexes[-1].Point)
            return Part.LineSegment(v1, v2).toShape()
        return edge

    def flattenFace(self, face):
        "Returns a face where all vertices have Z = 0"
        wires = []
        for w in face[0].Wires:
            verts = []
            edges = Part.__sortEdges__(w.Edges)
            for e in edges:
                v = e.Vertexes[0].Point
                verts.append(FreeCAD.Vector(v.x, v.y, 0))
            verts.append(verts[0])
            wires.append(Part.makePolygon(verts))
        try:
            sh = Part.Face(wires)
        except Part.OCCError:
            if DEBUG:
                print("Error: Unable to flatten face")
            return None
        else:
            return [sh] + face[1:]

    def cut(self, cutplane, hidden=False):
        "Cuts through the shapes with a given cut plane and builds section faces"
        if DEBUG:
            print("\n\n======> Starting cut\n\n")
        if self.iscut:
            return
        if not self.shapes:
            if DEBUG:
                print("No objects to make sections")
        else:
            fill = (1.0, 1.0, 1.0, 1.0)
            shps = []
            for sh in self.shapes:
                shps.append(sh[0])
            cutface, cutvolume, invcutvolume = ArchCommands.getCutVolume(cutplane, shps)
            if cutface and cutvolume:
                shapes = []
                faces = []
                sections = []
                for sh in self.shapes:
                    for sol in sh[0].Solids:
                        c = sol.cut(cutvolume)
                        shapes.append([c] + sh[1:])
                        for f in c.Faces:
                            faces.append([f] + sh[1:])
                            # print("iscoplanar:",f.Vertexes[0].Point,f.normalAt(0,0),cutface.Vertexes[0].Point,cutface.normalAt(0,0))
                            if DraftGeomUtils.isCoplanar([f, cutface]):
                                print("COPLANAR")
                                sections.append([f, fill])
                        if hidden:
                            c = sol.cut(invcutvolume)
                            self.hiddenEdges.extend(c.Edges)
                self.shapes = shapes
                self.faces = faces
                self.sections = sections
                if DEBUG:
                    print(
                        "Built ",
                        len(self.sections),
                        " sections, ",
                        len(self.faces),
                        " faces retained",
                    )
                self.iscut = True
                self.oriented = False
                self.trimmed = False
                self.sorted = False
                self.joined = False
        if DEBUG:
            print("\n\n======> Finished cut\n\n")

    def isInside(self, vert, face):
        "Returns True if the vert is inside the face in Z projection"

        if not face:
            return False

        # http://paulbourke.net/geometry/insidepoly/
        count = 0
        p = self.wp.get_local_coords(vert.Point)
        for e in face[0].Edges:
            p1 = e.Vertexes[0].Point
            p2 = e.Vertexes[-1].Point
            if p.y > min(p1.y, p2.y):
                if p.y <= max(p1.y, p2.y):
                    if p.x <= max(p1.x, p2.x):
                        if p1.y != p2.y:
                            xinters = (p.y - p1.y) * (p2.x - p1.x) / (p2.y - p1.y) + p1.x
                            if (p1.x == p2.x) or (p.x <= xinters):
                                count += 1
        if count % 2 == 0:
            return False
        else:
            return True

    def zOverlaps(self, face1, face2):
        "Checks if face1 overlaps face2 in Z direction"
        face1 = self.flattenFace(face1)
        face2 = self.flattenFace(face2)

        if (not face1) or (not face2):
            return False

        # first we check if one of the verts is inside the other face
        for v in face1[0].Vertexes:
            if self.isInside(v, face2):
                return True

        # even so, faces can still overlap if their edges cross each other
        for e1 in face1[0].Edges:
            for e2 in face2[0].Edges:
                if DraftGeomUtils.findIntersection(e1, e2):
                    return True
        return False

    def compare(self, face1, face2):
        "zsorts two faces. Returns 1 if face1 is closer, 2 if face2 is closer, 0 otherwise"

        # print(face1,face2)

        if not face1:
            if DEBUG:
                print("Warning, undefined face!")
            return 31
        elif not face2:
            if DEBUG:
                print("Warning, undefined face!")
            return 32

        # theory from
        # http://www.siggraph.org/education/materials/HyperGraph/scanline/visibility/painter.htm
        # and practical application http://vrm.ao2.it/ (blender vector renderer)

        b1 = face1[0].BoundBox
        b2 = face2[0].BoundBox

        # test 1: if faces don't overlap, no comparison possible
        if DEBUG:
            print("doing test 1")
        if b1.XMax < b2.XMin:
            return 0
        if b1.XMin > b2.XMax:
            return 0
        if b1.YMax < b2.YMin:
            return 0
        if b1.YMin > b2.YMax:
            return 0
        if DEBUG:
            print("failed, faces bboxes are not distinct")

        # test 2: if Z bounds don't overlap, it's easy to know the closest
        if DEBUG:
            print("doing test 2")
        if b1.ZMax < b2.ZMin:
            return 2
        if b2.ZMax < b1.ZMin:
            return 1
        if DEBUG:
            print("failed, faces Z are not distinct")

        # test 3: all verts of face1 are in front or behind the plane of face2
        if DEBUG:
            print("doing test 3")
        norm = face2[0].normalAt(0, 0)
        behind = 0
        front = 0
        for v in face1[0].Vertexes:
            dv = v.Point.sub(face2[0].Vertexes[0].Point)
            dv = DraftVecUtils.project(dv, norm)
            if DraftVecUtils.isNull(dv):
                behind += 1
                front += 1
            else:
                if dv.getAngle(norm) > 1:
                    behind += 1
                else:
                    front += 1
        if DEBUG:
            print("front: ", front, " behind: ", behind)
        if behind == len(face1[0].Vertexes):
            return 2
        elif front == len(face1[0].Vertexes):
            return 1
        if DEBUG:
            print("failed, cannot say if face 1 is in front or behind")

        # test 4: all verts of face2 are in front or behind the plane of face1
        if DEBUG:
            print("doing test 4")
        norm = face1[0].normalAt(0, 0)
        behind = 0
        front = 0
        for v in face2[0].Vertexes:
            dv = v.Point.sub(face1[0].Vertexes[0].Point)
            dv = DraftVecUtils.project(dv, norm)
            if DraftVecUtils.isNull(dv):
                behind += 1
                front += 1
            else:
                if dv.getAngle(norm) > 1:
                    behind += 1
                else:
                    front += 1
        if DEBUG:
            print("front: ", front, " behind: ", behind)
        if behind == len(face2[0].Vertexes):
            return 1
        elif front == len(face2[0].Vertexes):
            return 2
        if DEBUG:
            print("failed, cannot say if face 2 is in front or behind")

        # test 5: see if faces projections don't overlap, vertexwise
        if DEBUG:
            print("doing test 5")
        if not self.zOverlaps(face1, face2):
            return 0
        elif not self.zOverlaps(face2, face1):
            return 0
        if DEBUG:
            print("failed, faces are overlapping")

        if DEBUG:
            print("Houston, all tests passed, and still no results")
        return 0

    def join(self, otype):
        "joins the objects of same type"
        import Part

        walls = []
        structs = []
        objs = []
        for o in obj.Source.Objects:
            t = Draft.getType(o)
            if t == "Wall":
                walls.append(o)
            elif t == "Structure":
                structs.append(o)
            else:
                objs.append(o)
        for g in [walls, structs]:
            if g:
                print("group:", g)
                col = g[0].ViewObject.DiffuseColor[0]
                s = g[0].Shape
                for o in g[1:]:
                    try:
                        fs = s.fuse(o.Shape)
                        fs = fs.removeSplitter()
                    except Part.OCCError:
                        print("shape fusion failed")
                        objs.append([o.Shape, o.ViewObject.DiffuseColor[0]])
                    else:
                        s = fs
                objs.append([s, col])

    def findPosition(self, f1, faces):
        "Finds the position of a face in a list of faces"
        l = None
        h = None
        for f2 in faces:
            if DEBUG:
                print(
                    "comparing face",
                    str(self.faces.index(f1)),
                    " with face",
                    str(self.faces.index(f2)),
                )
            r = self.compare(f1, f2)
            if r == 1:
                l = faces.index(f2)
            elif r == 2:
                if h is None:
                    h = faces.index(f2)
                else:
                    if faces.index(f2) < h:
                        h = faces.index(f2)
        if l is not None:
            return l + 1
        elif h is not None:
            return h
        else:
            return None

    def sort(self):
        "projects a shape on the WP"
        if DEBUG:
            print("\n\n======> Starting sort\n\n")
        if len(self.faces) <= 1:
            return
        if not self.trimmed:
            self.removeHidden()
            if DEBUG:
                print("Done hidden face removal")
        if len(self.faces) == 1:
            return
        if not self.oriented:
            self.reorient()
            if DEBUG:
                print("Done reorientation")
        faces = self.faces[:]
        if DEBUG:
            print("sorting ", len(self.faces), " faces")
        sfaces = []
        loopcount = 0
        notfoundstack = 0
        while faces:
            if DEBUG:
                print("loop ", loopcount)
            f1 = faces[0]
            if sfaces and (notfoundstack < len(faces)):
                if DEBUG:
                    print("using ordered stack, notfound = ", notfoundstack)
                p = self.findPosition(f1, sfaces)
                if p is None:
                    # no position found, we move the face to the end of the pile
                    faces.remove(f1)
                    faces.append(f1)
                    notfoundstack += 1
                else:
                    # position found, we insert it
                    faces.remove(f1)
                    sfaces.insert(p, f1)
                    notfoundstack = 0
            else:
                # either there is no stack, or no more face can be compared
                # find a root, 2 faces that can be compared
                if DEBUG:
                    print("using unordered stack, notfound = ", notfoundstack)
                for f2 in faces[1:]:
                    if DEBUG:
                        print(
                            "comparing face",
                            str(self.faces.index(f1)),
                            " with face",
                            str(self.faces.index(f2)),
                        )
                    r = self.compare(f1, f2)
                    print("comparison result:", r)
                    if r == 1:
                        faces.remove(f2)
                        sfaces.append(f2)
                        faces.remove(f1)
                        sfaces.append(f1)
                        notfoundstack = 0
                        break
                    elif r == 2:
                        faces.remove(f1)
                        sfaces.append(f1)
                        faces.remove(f2)
                        sfaces.append(f2)
                        notfoundstack = 0
                        break
                    elif r == 31:
                        if f1 in faces:
                            faces.remove(f1)
                    elif r == 32:
                        if f2 in faces:
                            faces.remove(f2)
                else:
                    # nothing found, move the face to the end of the pile
                    if f1 in faces:
                        faces.remove(f1)
                        faces.append(f1)
            loopcount += 1
            if loopcount == MAXLOOP * len(self.faces):
                if DEBUG:
                    print("Too many loops, aborting.")
                break

        if DEBUG:
            print(
                "done Z sorting. ",
                len(sfaces),
                " faces retained, ",
                len(self.faces) - len(sfaces),
                " faces lost.",
            )
        self.faces = sfaces
        self.sorted = True
        if DEBUG:
            print("\n\n======> Finished sort\n\n")

    def buildDummy(self):
        "Builds a dummy object with faces spaced on the Z axis, for visual check"
        z = 0
        if not self.sorted:
            self.sort()
        faces = []
        for f in self.faces[:]:
            ff = self.flattenFace(f)[0]
            ff.translate(FreeCAD.Vector(0, 0, z))
            faces.append(ff)
            z += 1
        if faces:
            c = Part.makeCompound(faces)
            Part.show(c)

    def getFill(self, fill):
        "Returns a SVG fill value"
        r = str(hex(int(fill[0] * 255)))[2:].zfill(2)
        g = str(hex(int(fill[1] * 255)))[2:].zfill(2)
        b = str(hex(int(fill[2] * 255)))[2:].zfill(2)
        col = "#" + r + g + b
        return col

    def getPathData(self, w):
        "Returns a SVG path data string from a 2D wire"

        def tostr(val):
            return str(round(val, Draft.precision()))

        edges = Part.__sortEdges__(w.Edges)
        v = edges[0].Vertexes[0].Point
        svg = "M " + tostr(v.x) + " " + tostr(v.y) + " "
        for e in edges:
            if (DraftGeomUtils.geomType(e) == "Line") or (
                DraftGeomUtils.geomType(e) == "BSplineCurve"
            ):
                v = e.Vertexes[-1].Point
                svg += "L " + tostr(v.x) + " " + tostr(v.y) + " "
            elif DraftGeomUtils.geomType(e) == "Circle":
                r = e.Curve.Radius
                v = e.Vertexes[-1].Point
                svg += "A " + tostr(r) + " " + tostr(r) + " 0 0 1 " + tostr(v.x) + " "
                svg += tostr(v.y) + " "
        if len(edges) > 1:
            svg += "Z "
        return svg

    def getViewSVG(self, linewidth=0.01):
        "Returns a SVG fragment from viewed faces"
        if DEBUG:
            print("Printing ", len(self.faces), " faces")
        if not self.sorted:
            self.sort()
        svg = (
            '<g stroke="#000000" stroke-width="'
            + str(linewidth)
            + '" style="stroke-width:'
            + str(linewidth)
        )
        svg += ';stroke-miterlimit:1;stroke-linejoin:round;stroke-dasharray:none;">\n'
        for f in self.faces:
            if f:
                fill = self.getFill(f[1])
                svg += "    <path "
                svg += 'd="'
                for w in f[0].Wires:
                    svg += self.getPathData(w)
                svg += '" style="fill:' + fill + ';fill-rule: evenodd;"/>\n'
        svg += "</g>\n"
        return svg

    def getSectionSVG(self, linewidth=0.02, fillpattern=None):
        "Returns a SVG fragment from cut faces"
        if DEBUG:
            print("Printing ", len(self.sections), " sections")
        if not self.oriented:
            self.reorient()
        svg = (
            '<g stroke="#000000" stroke-width="'
            + str(linewidth)
            + '" style="stroke-width:'
            + str(linewidth)
        )
        svg += ';stroke-miterlimit:1;stroke-linejoin:round;stroke-dasharray:none;">\n'
        for f in self.sections:
            if f:
                if fillpattern:
                    if "#" in fillpattern:  # color
                        fill = fillpattern
                    else:
                        fill = "url(#" + fillpattern + ")"  # pattern name
                else:
                    fill = "none"  # none
                svg += "<path "
                svg += 'd="'
                for w in f[0].Wires:
                    # print("wire with ",len(w.Vertexes)," verts")
                    svg += self.getPathData(w)
                svg += '" style="fill:' + fill + ';fill-rule: evenodd;"/>\n'
        svg += "</g>\n"
        return svg

    def getHiddenSVG(self, linewidth=0.02):
        "Returns a SVG fragment from cut geometry"
        if DEBUG:
            print("Printing ", len(self.sections), " hidden faces")
        if not self.oriented:
            self.reorient()
        svg = (
            '<g stroke="#000000" stroke-width="'
            + str(linewidth)
            + '" style="stroke-width:'
            + str(linewidth)
        )
        svg += (
            ';stroke-miterlimit:1;stroke-linejoin:round;stroke-dasharray:0.09,0.05;fill:none;">\n'
        )
        for e in self.hiddenEdges:
            svg += "<path "
            svg += 'd="'
            svg += self.getPathData(e)
            svg += '"/>\n'
        svg += "</g>\n"
        return svg



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the task panel code for the Draft CircularArray tool."""
## @package task_circulararray
# \ingroup drafttaskpanels
# \brief Provides the task panel code for the Draft CircularArray tool.

## \addtogroup drafttaskpanels
# @{
import PySide.QtGui as QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc  # include resources, icons, ui files
import DraftVecUtils
from FreeCAD import Units as U
from draftutils import params
from draftutils.messages import _err, _log, _msg, _wrn
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
bool(Draft_rc.__name__)


def _quantity(st):
    # workaround for improper handling of plus sign
    # in Building US unit system
    # https://github.com/FreeCAD/FreeCAD/issues/11345
    return U.Quantity(st.replace("+", "--")).Value


class TaskPanelCircularArray:
    """TaskPanel code for the CircularArray command.

    The names of the widgets are defined in the `.ui` file.
    This `.ui` file `must` be loaded into an attribute
    called `self.form` so that it is loaded into the task panel correctly.

    In this class all widgets are automatically created
    as `self.form.<widget_name>`.

    The `.ui` file may use special FreeCAD widgets such as
    `Gui::InputField` (based on `QLineEdit`) and
    `Gui::QuantitySpinBox` (based on `QAbstractSpinBox`).
    See the Doxygen documentation of the corresponding files in `src/Gui/`,
    for example, `InputField.h` and `QuantitySpinBox.h`.

    Attributes
    ----------
    source_command: gui_base.GuiCommandBase
        This attribute holds a reference to the calling class
        of this task panel.
        This parent class, which is derived from `gui_base.GuiCommandBase`,
        is responsible for calling this task panel, for installing
        certain callbacks, and for removing them.

        It also delays the execution of the internal creation commands
        by using the `draftutils.todo.ToDo` class.

    See Also
    --------
    * https://forum.freecad.org/viewtopic.php?f=10&t=40007
    * https://forum.freecad.org/viewtopic.php?t=5374#p43038
    """

    def __init__(self):

        self.form = Gui.PySideUic.loadUi(":/ui/TaskPanel_CircularArray.ui")
        self.form.setWindowTitle(translate("draft", "Circular Array"))
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_CircularArray.svg"))

        # -------------------------------------------------------------------
        # Default values for the internal function, and for the task panel interface
        self.center = App.Vector()
        # TODO: the axis is currently fixed, it should be editable
        # or selectable from the task panel
        self.axis = App.Vector(0, 0, 1)
        self.r_distance = 100
        self.tan_distance = 50
        self.number = 3
        self.symmetry = 1
        self.fuse = params.get_param("Draft_array_fuse")
        self.use_link = params.get_param("Draft_array_Link")

        self.form.input_c_x.setProperty("rawValue", self.center.x)
        self.form.input_c_y.setProperty("rawValue", self.center.y)
        self.form.input_c_z.setProperty("rawValue", self.center.z)
        self.form.spinbox_r_distance.setProperty("rawValue", self.r_distance)
        self.form.spinbox_tan_distance.setProperty("rawValue", self.tan_distance)
        self.form.spinbox_number.setValue(self.number)
        self.form.spinbox_symmetry.setValue(self.symmetry)
        self.form.checkbox_fuse.setChecked(self.fuse)
        self.form.checkbox_link.setChecked(self.use_link)
        # -------------------------------------------------------------------

        # Some objects need to be selected before we can execute the function.
        self.selection = None

        # This is used to test the input of the internal function.
        # It should be changed to True before we can execute the function.
        self.valid_input = False

        self.set_widget_callbacks()

        self.tr_true = QT_TRANSLATE_NOOP("Draft", "True")
        self.tr_false = QT_TRANSLATE_NOOP("Draft", "False")

        # The mask is not used at the moment, but could be used in the future
        # by a callback to restrict the coordinates of the pointer.
        self.mask = ""

    def set_widget_callbacks(self):
        """Set up the callbacks (slots) for the widget signals."""
        # New style for Qt5
        self.form.button_reset.clicked.connect(self.reset_point)

        # When the checkbox changes, change the internal value
        if hasattr(self.form.checkbox_fuse, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.checkbox_fuse.checkStateChanged.connect(self.set_fuse)
            self.form.checkbox_link.checkStateChanged.connect(self.set_link)
        else:  # Qt version < 6.7.0
            self.form.checkbox_fuse.stateChanged.connect(self.set_fuse)
            self.form.checkbox_link.stateChanged.connect(self.set_link)

    def accept(self):
        """Execute when clicking the OK button or Enter key."""
        self.selection = Gui.Selection.getSelection()

        (self.r_distance, self.tan_distance) = self.get_distances()

        (self.number, self.symmetry) = self.get_number_symmetry()

        self.axis = self.get_axis()
        self.center = self.get_center()

        self.valid_input = self.validate_input(
            self.selection,
            self.r_distance,
            self.tan_distance,
            self.number,
            self.symmetry,
            self.axis,
            self.center,
        )
        if self.valid_input:
            self.create_object()
            # The internal function already displays messages
            self.finish()

    def validate_input(self, selection, r_distance, tan_distance, number, symmetry, axis, center):
        """Check that the input is valid.

        Some values may not need to be checked because
        the interface may not allow one to input wrong data.
        """
        if not selection:
            _err(translate("draft", "At least 1 element must be selected"))
            return False

        if number < 2:
            _err(translate("draft", "Number of layers must be at least 2"))
            return False

        # TODO: this should handle multiple objects.
        # Each of the elements of the selection should be tested.
        obj = selection[0]
        if obj.isDerivedFrom("App::FeaturePython"):
            _err(translate("draft", "Selection is not suitable for array"))
            _err(translate("draft", "Object:") + " {}".format(selection[0].Label))
            return False

        if r_distance == 0:
            _wrn(
                translate("draft", "Radial distance is zero. Resulting array may not look correct.")
            )
        elif r_distance < 0:
            _wrn(translate("draft", "Radial distance is negative. It is made positive to proceed."))
            self.r_distance = abs(r_distance)

        if tan_distance == 0:
            _err(translate("draft", "Tangential distance cannot be 0"))
            return False
        elif tan_distance < 0:
            _wrn(
                translate(
                    "draft", "Tangential distance is negative. It is made positive to proceed."
                )
            )
            self.tan_distance = abs(tan_distance)

        # The other arguments are not tested but they should be present.
        if symmetry and axis and center:
            pass

        self.fuse = self.form.checkbox_fuse.isChecked()
        self.use_link = self.form.checkbox_link.isChecked()
        return True

    def create_object(self):
        """Create the new object.

        At this stage we already tested that the input is correct
        so the necessary attributes are already set.
        Then we proceed with the internal function to create the new object.
        """
        if len(self.selection) == 1:
            sel_obj = self.selection[0]
        else:
            # TODO: this should handle multiple objects.
            # For example, it could take the shapes of all objects,
            # make a compound and then use it as input for the array function.
            sel_obj = self.selection[0]

        # This creates the object immediately
        # obj = Draft.make_circular_array(sel_obj,
        #                                 self.r_distance, self.tan_distance,
        #                                 self.number, self.symmetry,
        #                                 self.axis, self.center,
        #                                 self.use_link)

        # Instead, we build the commands to execute through the caller
        # of this class, the GuiCommand.
        # This is needed to schedule geometry manipulation
        # that would crash Coin3D if done in the event callback.
        _cmd = "Draft.make_circular_array"
        _cmd += "("
        _cmd += "App.ActiveDocument." + sel_obj.Name + ", "
        _cmd += "r_distance=" + str(self.r_distance) + ", "
        _cmd += "tan_distance=" + str(self.tan_distance) + ", "
        _cmd += "number=" + str(self.number) + ", "
        _cmd += "symmetry=" + str(self.symmetry) + ", "
        _cmd += "axis=" + DraftVecUtils.toString(self.axis) + ", "
        _cmd += "center=" + DraftVecUtils.toString(self.center) + ", "
        _cmd += "use_link=" + str(self.use_link)
        _cmd += ")"

        Gui.addModule("Draft")

        _cmd_list = [
            "_obj_ = " + _cmd,
            "_obj_.Fuse = " + str(self.fuse),
            "Draft.autogroup(_obj_)",
            "App.ActiveDocument.recompute()",
        ]

        # We commit the command list through the parent command
        self.source_command.commit(translate("draft", "Create Circular Array"), _cmd_list)

    def get_distances(self):
        """Get the distance parameters from the widgets."""
        r_d_str = self.form.spinbox_r_distance.text()
        tan_d_str = self.form.spinbox_tan_distance.text()
        return _quantity(r_d_str), _quantity(tan_d_str)

    def get_number_symmetry(self):
        """Get the number and symmetry parameters from the widgets."""
        number = self.form.spinbox_number.value()
        symmetry = self.form.spinbox_symmetry.value()
        return number, symmetry

    def get_center(self):
        """Get the value of the center from the widgets."""
        c_x_str = self.form.input_c_x.text()
        c_y_str = self.form.input_c_y.text()
        c_z_str = self.form.input_c_z.text()
        center = App.Vector(_quantity(c_x_str), _quantity(c_y_str), _quantity(c_z_str))
        return center

    def get_axis(self):
        """Get the axis that will be used for the array. NOT IMPLEMENTED.

        It should consider a second selection of an edge or wire to use
        as an axis.
        """
        return self.axis

    def reset_point(self):
        """Reset the center point to the original distance."""
        self.form.input_c_x.setProperty("rawValue", 0)
        self.form.input_c_y.setProperty("rawValue", 0)
        self.form.input_c_z.setProperty("rawValue", 0)

        self.center = self.get_center()

    def print_fuse_state(self, fuse):
        """Print the fuse state translated."""
        if fuse:
            state = self.tr_true
        else:
            state = self.tr_false
        _msg(translate("draft", "Fuse:") + " {}".format(state))

    def set_fuse(self):
        """Execute as a callback when the fuse checkbox changes."""
        self.fuse = self.form.checkbox_fuse.isChecked()
        params.set_param("Draft_array_fuse", self.fuse)

    def print_link_state(self, use_link):
        """Print the link state translated."""
        if use_link:
            state = self.tr_true
        else:
            state = self.tr_false
        _msg(translate("draft", "Create Link array:") + " {}".format(state))

    def set_link(self):
        """Execute as a callback when the link checkbox changes."""
        self.use_link = self.form.checkbox_link.isChecked()
        params.set_param("Draft_array_Link", self.use_link)

    def print_messages(self):
        """Print messages about the operation."""
        if len(self.selection) == 1:
            sel_obj = self.selection[0]
        else:
            # TODO: this should handle multiple objects.
            # For example, it could take the shapes of all objects,
            # make a compound and then use it as input for the array function.
            sel_obj = self.selection[0]
        _msg(translate("draft", "Object:") + " {}".format(sel_obj.Label))
        _msg(translate("draft", "Radial distance:") + " {}".format(self.r_distance))
        _msg(translate("draft", "Tangential distance:") + " {}".format(self.tan_distance))
        _msg(translate("draft", "Number of concentric circles:") + " {}".format(self.number))
        _msg(translate("draft", "Symmetry parameter:") + " {}".format(self.symmetry))
        _msg(
            translate("draft", "Center of rotation:")
            + " ({0}, {1}, {2})".format(self.center.x, self.center.y, self.center.z)
        )
        self.print_fuse_state(self.fuse)
        self.print_link_state(self.use_link)

    def display_point(self, point=None, plane=None, mask=None):
        """Display the coordinates in the x, y, and z widgets.

        This function should be used in a Coin callback so that
        the coordinate values are automatically updated when the
        mouse pointer moves.
        This was copied from `DraftGui.py` but needs to be improved
        for this particular command.

        point: Base::Vector3
            is a vector that arrives by the callback.
        plane: WorkingPlane.PlaneGui
            is a working plane instance. Not used at the moment.
        mask: str
            is a string that specifies which coordinate is being
            edited. It is used to restrict edition of a single coordinate.
            It is not used at the moment but could be used with a callback.
        """
        # Get the coordinates to display
        d_p = None
        if point:
            d_p = point

        # Set the widgets to the value of the mouse pointer.
        #
        # setProperty() is used if the widget is a FreeCAD widget like
        # Gui::InputField or Gui::QuantitySpinBox, which are based on
        # QLineEdit and QAbstractSpinBox.
        #
        # setText() is used to set the text inside the widget, this may be
        # useful in some circumstances.
        #
        # The mask allows editing only one field, that is, only one coordinate.
        # sbx = self.form.spinbox_c_x
        # sby = self.form.spinbox_c_y
        # sbz = self.form.spinbox_c_z
        if d_p:
            if self.mask in ("y", "z"):
                # sbx.setText(displayExternal(d_p.x, None, 'Length'))
                self.form.input_c_x.setProperty("rawValue", d_p.x)
            else:
                # sbx.setText(displayExternal(d_p.x, None, 'Length'))
                self.form.input_c_x.setProperty("rawValue", d_p.x)
            if self.mask in ("x", "z"):
                # sby.setText(displayExternal(d_p.y, None, 'Length'))
                self.form.input_c_y.setProperty("rawValue", d_p.y)
            else:
                # sby.setText(displayExternal(d_p.y, None, 'Length'))
                self.form.input_c_y.setProperty("rawValue", d_p.y)
            if self.mask in ("x", "y"):
                # sbz.setText(displayExternal(d_p.z, None, 'Length'))
                self.form.input_c_z.setProperty("rawValue", d_p.z)
            else:
                # sbz.setText(displayExternal(d_p.z, None, 'Length'))
                self.form.input_c_z.setProperty("rawValue", d_p.z)

        if plane:
            pass

        # Set masks
        if (mask == "x") or (self.mask == "x"):
            self.form.input_c_x.setEnabled(True)
            self.form.input_c_y.setEnabled(False)
            self.form.input_c_z.setEnabled(False)
            self.set_focus("x")
        elif (mask == "y") or (self.mask == "y"):
            self.form.input_c_x.setEnabled(False)
            self.form.input_c_y.setEnabled(True)
            self.form.input_c_z.setEnabled(False)
            self.set_focus("y")
        elif (mask == "z") or (self.mask == "z"):
            self.form.input_c_x.setEnabled(False)
            self.form.input_c_y.setEnabled(False)
            self.form.input_c_z.setEnabled(True)
            self.set_focus("z")
        else:
            self.form.input_c_x.setEnabled(True)
            self.form.input_c_y.setEnabled(True)
            self.form.input_c_z.setEnabled(True)
            self.set_focus()

    def set_focus(self, key=None):
        """Set the focus on the widget that receives the key signal."""
        if key is None or key == "x":
            self.form.input_c_x.setFocus()
            self.form.input_c_x.selectAll()
        elif key == "y":
            self.form.input_c_y.setFocus()
            self.form.input_c_y.selectAll()
        elif key == "z":
            self.form.input_c_z.setFocus()
            self.form.input_c_z.selectAll()

    def reject(self):
        """Execute when clicking the Cancel button or pressing Escape."""
        self.finish()

    def finish(self):
        """Finish the command, after accept or reject.

        It finally calls the parent class to execute
        the delayed functions, and perform cleanup.
        """
        # App.ActiveDocument.commitTransaction()
        if Gui.ActiveDocument is not None:
            Gui.ActiveDocument.resetEdit()
        # Runs the parent command to complete the call
        self.source_command.completed()


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2016 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD 3DS importer"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package import3DS
#  \ingroup ARCH
#  \brief 3DS file format importer
#
#  This module provides tools to import 3DS files.

import os

import FreeCAD
import Mesh

DEBUG = True


def check3DS():
    "checks if collada if available"
    global dom3ds
    dom3ds = None
    try:
        from Dice3DS import dom3ds
    except ImportError:
        FreeCAD.Console.PrintError("Dice3DS not found, 3DS support is disabled.\n")
        return False
    else:
        return True


def open(filename):
    "called when freecad wants to open a file"
    if not check3DS():
        return
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    FreeCAD.ActiveDocument = doc
    read(filename)
    return doc


def insert(filename, docname):
    "called when freecad wants to import a file"
    if not check3DS():
        return
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc
    read(filename)
    return doc


def read(filename):
    dom = dom3ds.read_3ds_file(filename, tight=False)

    for j, d_nobj in enumerate(dom.mdata.objects):
        if type(d_nobj.obj) != dom3ds.N_TRI_OBJECT:
            continue
        verts = []
        if d_nobj.obj.points:
            for d_point in d_nobj.obj.points.array:
                verts.append([d_point[0], d_point[1], d_point[2]])
            meshdata = []
            for d_face in d_nobj.obj.faces.array:
                meshdata.append([verts[int(d_face[i])] for i in range(3)])
            m = [tuple(r) for r in d_nobj.obj.matrix.array]
            m = m[0] + m[1] + m[2] + m[3]
            placement = FreeCAD.Placement(FreeCAD.Matrix(*m))
            mesh = Mesh.Mesh(meshdata)
            obj = FreeCAD.ActiveDocument.addObject("Mesh::Feature", "Mesh")
            obj.Mesh = mesh
            obj.Placement = placement
        else:
            print("Skipping object without vertices array: ", d_nobj.obj)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import uuid
import json
from typing import Mapping, List, Optional, Type
import FreeCAD
from ...assets import Asset, AssetUri, AssetSerializer
from ...toolbit import ToolBit
from ...toolbit.mixins import RotaryToolBitMixin
from ...shape import ToolBitShape, ToolBitShapeEndmill
from ..models.library import Library

SHAPEMAP = {
    "ballend": "Ballnose",
    "endmill": "Cylindrical",
    "taperedballnose": "Ballnose",
    "v-bit": "Conical",
    "vbit": "Conical",
    "chamfer": "Snubnose",
}

SHAPEMAP_REVERSE = {
    "Ballnose": "ballend",  # Default to ballend when deserializing Ballnose
    "Cylindrical": "endmill",
    "Conical": "v-bit",
    "Snubnose": "chamfer",
}

tooltemplate = {
    "units": "metric",
    "shape": "Cylindrical",
    "length": 10,
    "diameter": 3.125,
    "description": "",
}


class CamoticsLibrarySerializer(AssetSerializer):
    for_class: Type[Asset] = Library
    extensions: tuple[str] = (".json",)
    mime_type: str = "application/json"

    @classmethod
    def get_label(cls) -> str:
        return FreeCAD.Qt.translate("CAM", "Camotics Tool Library")

    @classmethod
    def extract_dependencies(cls, data: bytes) -> List[AssetUri]:
        return []

    @classmethod
    def serialize(cls, asset: Asset) -> bytes:
        if not isinstance(asset, Library):
            raise TypeError("Asset must be a Library instance")

        toollist = {}
        for tool_no, tool in asset._bit_nos.items():
            assert isinstance(tool, RotaryToolBitMixin)
            toolitem = tooltemplate.copy()

            diameter_value = tool.get_diameter()
            # Ensure diameter is a float, handle Quantity and other types
            diameter_serializable = 2.0  # Default value as float
            if isinstance(diameter_value, FreeCAD.Units.Quantity):
                try:
                    val_mm = diameter_value.getValueAs("mm")
                    if val_mm is not None:
                        diameter_serializable = float(val_mm)
                except ValueError:
                    # Fallback to raw value if unit conversion fails
                    raw_val = diameter_value.Value if hasattr(diameter_value, "Value") else None
                    if isinstance(raw_val, (int, float)):
                        diameter_serializable = float(raw_val)
            elif isinstance(diameter_value, (int, float)):
                diameter_serializable = float(diameter_value) if diameter_value is not None else 2.0

            toolitem["diameter"] = diameter_serializable

            toolitem["description"] = tool.label

            length_value = tool.get_length()
            # Ensure length is a float, handle Quantity and other types
            length_serializable = 10.0  # Default value as float
            if isinstance(length_value, FreeCAD.Units.Quantity):
                try:
                    val_mm = length_value.getValueAs("mm")
                    if val_mm is not None:
                        length_serializable = float(val_mm)
                except ValueError:
                    # Fallback to raw value if unit conversion fails
                    raw_val = length_value.Value if hasattr(length_value, "Value") else None
                    if isinstance(raw_val, (int, float)):
                        length_serializable = float(raw_val)
            elif isinstance(length_value, (int, float)):
                length_serializable = float(length_value) if length_value is not None else 10.0

            toolitem["length"] = length_serializable

            toolitem["shape"] = SHAPEMAP.get(tool._tool_bit_shape.name.lower(), "Cylindrical")
            toollist[str(tool_no)] = toolitem

        return json.dumps(toollist, indent=2).encode("utf-8")

    @classmethod
    def deserialize(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> Library:
        try:
            data_dict = json.loads(data.decode("utf-8"))
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to decode JSON data: {e}") from e

        library = Library(id, id=id)
        for tool_no_str, toolitem in data_dict.items():
            try:
                tool_no = int(tool_no_str)
            except ValueError:
                print(f"Warning: Skipping invalid tool number: {tool_no_str}")
                continue

            # Find the shape class to use
            shape_name_str = SHAPEMAP_REVERSE.get(toolitem.get("shape", "Cylindrical"), "endmill")
            shape_class = ToolBitShape.get_subclass_by_name(shape_name_str)
            if not shape_class:
                print(f"Warning: Unknown shape name '{shape_name_str}', defaulting to endmill")
                shape_class = ToolBitShapeEndmill

            # Translate parameters to FreeCAD types
            params = {}
            try:
                diameter = float(toolitem.get("diameter", 2))
                params["Diameter"] = FreeCAD.Units.Quantity(f"{diameter} mm")
            except (ValueError, TypeError):
                print(f"Warning: Invalid diameter for tool {tool_no_str}, skipping.")

            try:
                length = float(toolitem.get("length", 10))
                params["Length"] = FreeCAD.Units.Quantity(f"{length} mm")
            except (ValueError, TypeError):
                print(f"Warning: Invalid length for tool {tool_no_str}, skipping.")

            # Create the shape
            shape_id = shape_name_str.lower()
            tool_bit_shape = shape_class(shape_id, **params)

            # Create the toolbit
            tool = ToolBit(tool_bit_shape, id=f"camotics_tool_{tool_no_str}")
            tool.label = toolitem.get("description", "")

            library.add_bit(tool, tool_no)

        return library

    @classmethod
    def deep_deserialize(cls, data: bytes) -> Library:
        # TODO: Build tools here
        return cls.deserialize(data, str(uuid.uuid4()), {})



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2021 Russell Johnson (russ4262) <russ4262@gmail.com>    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path.Op.Adaptive as PathAdaptive
import Path.Main.Job as PathJob
from CAMTests.PathTestUtils import PathTestBase

if FreeCAD.GuiUp:
    import Path.Main.Gui.Job as PathJobGui
    import Path.Op.Gui.Adaptive as PathAdaptiveGui


class TestPathAdaptive(PathTestBase):
    """Unit tests for the Adaptive operation."""

    @classmethod
    def setUpClass(cls):
        """setUpClass()...
        This method is called upon instantiation of this test class.  Add code and objects here
        that are needed for the duration of the test() methods in this class.  In other words,
        set up the 'global' test environment here; use the `setUp()` method to set up a 'local'
        test environment.
        This method does not have access to the class `self` reference, but it
        is able to call static methods within this same class.
        """
        cls.needsInit = True

    @classmethod
    def initClass(cls):
        # Open existing FreeCAD document with test geometry
        cls.needsInit = False
        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "True")
        cls.doc = FreeCAD.open(FreeCAD.getHomePath() + "Mod/CAM/CAMTests/test_adaptive.fcstd")
        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "")

        # Create Job object, adding geometry objects from file opened above
        cls.job = PathJob.Create("Job", [cls.doc.Fusion], None)
        cls.job.GeometryTolerance.Value = 0.001
        if FreeCAD.GuiUp:
            cls.job.ViewObject.Proxy = PathJobGui.ViewProvider(cls.job.ViewObject)

        # Instantiate an Adaptive operation for querying available properties
        cls.prototype = PathAdaptive.Create("Adaptive")
        cls.prototype.Base = [(cls.doc.Fusion, ["Face3"])]
        cls.prototype.Label = "Prototype"
        _addViewProvider(cls.prototype)

        cls.doc.recompute()

    @classmethod
    def tearDownClass(cls):
        """tearDownClass()...
        This method is called prior to destruction of this test class.  Add code and objects here
        that cleanup the test environment after the test() methods in this class have been executed.
        This method does not have access to the class `self` reference.  This method
        is able to call static methods within this same class.
        """
        # FreeCAD.Console.PrintMessage("TestPathAdaptive.tearDownClass()\n")

        # Close geometry document without saving
        if not cls.needsInit:
            FreeCAD.closeDocument(cls.doc.Name)

    # Setup and tear down methods called before and after each unit test
    def setUp(self):
        """setUp()...
        This method is called prior to each `test()` method.  Add code and objects here
        that are needed for multiple `test()` methods.
        """
        if self.needsInit:
            self.initClass()

    def tearDown(self):
        """tearDown()...
        This method is called after each test() method. Add cleanup instructions here.
        Such cleanup instructions will likely undo those in the setUp() method.
        """
        pass

    # Unit tests
    def testFaceSingleSimple(self):
        """testFaceSingleSimple() Verify path generated on Face3."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3"])]  # (base, subs_list)
        adaptive.Label = "testFaceSingleSimple+"
        adaptive.Comment = "testFaceSingleSimple() Verify path generated on Face3."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        # moves = getGcodeMoves(adaptive.Path.Commands, includeRapids=False)
        # operationMoves = ";  ".join(moves)
        # FreeCAD.Console.PrintMessage("test00_moves: " + operationMoves + "\n")

        # self.assertTrue(expected_moves_test01 == operationMoves,
        #                "expected_moves_test01: {}\noperationMoves: {}".format(expected_moves_test01, operationMoves))
        self.assertTrue(len(adaptive.Path.Commands) > 100, "Command count not greater than 100.")

    def testFacesMergedDifferentZ(self):
        """testFacesMergedDifferentZ() Verify path generated on adjacent, combined
        Face3 and Face10.  The Z heights are different."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3", "Face10"])]  # (base, subs_list)
        adaptive.Label = "testFacesMergedDifferentZ+"
        adaptive.Comment = "testFacesMergedDifferentZ() Verify path generated on adjacent, combined Face3 and Face10. The Z heights are different. UseOutline = False"

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(len(adaptive.Path.Commands) > 100, "Command count not greater than 100.")

    def testFacesMergedDifferentZUseOutline(self):
        """testFacesMergedDifferentZUseOutline() Verify path generated on adjacent, combined Face3 and Face10.
        The Z heights are different."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3", "Face10"])]  # (base, subs_list)
        adaptive.Label = "testFacesMergedDifferentZUseOutline+"
        adaptive.Comment = "testFacesMergedDifferentZUseOutline() Verify path generated on adjacent, combined Face3 and Face10.  The Z heights are different. UseOutline = True."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = True
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(len(adaptive.Path.Commands) > 100, "Command count not greater than 100.")

    def testOutlineDifferentZDiscontinuousEdges(self):
        """testOutlineDifferentZDiscontinuous() Verify path generated non-closed edges with differing Z-heights that are closed with Z=1 projection: "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19"."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [
            (
                self.doc.Fusion,
                [
                    "Edge9",
                    "Edge2",
                    "Edge8",
                    "Edge15",
                    "Edge30",
                    "Edge31",
                    "Edge29",
                    "Edge19",
                ],
            )
        ]  # (base, subs_list)
        adaptive.Label = "testOutlineDifferentZDiscontinuous+"
        adaptive.Comment = 'testOutlineDifferentZDiscontinuous() Verify path generated non-closed edges with differing Z-heights that are closed with Z=1 projection: "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19".'

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(len(adaptive.Path.Commands) > 100, "Command count not greater than 100.")

    def testOutlineDifferentZContinuousEdges(self):
        """test05() Verify path generated closed wire with differing Z-heights: "Edge13", "Edge7", "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19"."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [
            (
                self.doc.Fusion,
                [
                    "Edge13",
                    "Edge7",
                    "Edge9",
                    "Edge2",
                    "Edge8",
                    "Edge15",
                    "Edge30",
                    "Edge31",
                    "Edge29",
                    "Edge19",
                ],
            )
        ]  # (base, subs_list)
        adaptive.Label = "testOutlineDifferentZContinuous+"
        adaptive.Comment = 'testOutlineDifferentZContinuous() Verify path generated closed wire with differing Z-heights: "Edge13", "Edge7", "Edge9", "Edge2", "Edge8", "Edge15", "Edge30", "Edge31", "Edge29", "Edge19".'

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(len(adaptive.Path.Commands) > 100, "Command count not greater than 100.")

    def testOutlineWithCutout(self):
        """testOutlineWithCutout() Verify path generated with outer and inner edge loops at same Z height: "Edge15", "Edge30", "Edge31", "Edge29", "Edge19", "Edge18", "Edge35", "Edge32", "Edge34", "Edge33"."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [
            (
                self.doc.Fusion,
                [
                    "Edge15",
                    "Edge30",
                    "Edge31",
                    "Edge29",
                    "Edge19",
                    "Edge18",
                    "Edge35",
                    "Edge32",
                    "Edge34",
                    "Edge33",
                ],
            )
        ]  # (base, subs_list)
        adaptive.Label = "testOutlineWithCutout+"
        adaptive.Comment = 'testOutlineWithCutout() Verify path generated with outer and inner edge loops at same Z height: "Edge15", "Edge30", "Edge31", "Edge29", "Edge19", "Edge18", "Edge35", "Edge32", "Edge34", "Edge33".'

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        # Check command count
        self.assertTrue(len(adaptive.Path.Commands) > 100, "Command count not greater than 100.")

        # Check if any paths originate inside inner hole of donut.  They should not.
        isInBox = False
        edges = [
            self.doc.Fusion.Shape.getElement(e) for e in ["Edge35", "Edge32", "Edge33", "Edge34"]
        ]
        square = Part.Wire(edges)
        sqrBB = square.BoundBox
        minPoint = FreeCAD.Vector(sqrBB.XMin, sqrBB.YMin, 0.0)
        maxPoint = FreeCAD.Vector(sqrBB.XMax, sqrBB.YMax, 0.0)
        for c in adaptive.Path.Commands:
            if pathOriginatesInBox(c, minPoint, maxPoint):
                isInBox = True
                break
        self.assertFalse(isInBox, "Paths originating within the inner hole.")

    def testFaceWithCutout(self):
        """testFaceWithCutout() Verify path generated on donut-shaped Face10."""

        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face10"])]  # (base, subs_list)
        adaptive.Label = "testFaceWithCutout+"
        adaptive.Comment = "testFaceWithCutout() Verify path generated on donut-shaped Face10."

        # Set additional operation properties
        # setDepthsAndHeights(adaptive)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            20.0  # Have to set expression to None before numerical value assignment
        )

        _addViewProvider(adaptive)
        self.doc.recompute()

        self.assertTrue(len(adaptive.Path.Commands) > 100, "Command count not greater than 100.")

        # Check if any paths originate inside inner hole of donut.  They should not.
        isInBox = False
        edges = [
            self.doc.Fusion.Shape.getElement(e) for e in ["Edge35", "Edge32", "Edge33", "Edge34"]
        ]
        square = Part.Wire(edges)
        sqrBB = square.BoundBox
        minPoint = FreeCAD.Vector(sqrBB.XMin, sqrBB.YMin, 0.0)
        maxPoint = FreeCAD.Vector(sqrBB.XMax, sqrBB.YMax, 0.0)
        for c in adaptive.Path.Commands:
            if pathOriginatesInBox(c, minPoint, maxPoint):
                isInBox = True
                break
        self.assertFalse(isInBox, "Paths originating within the inner hole.")

        # Set Adaptive op to only use the outline of the face.
        adaptive.UseOutline = True
        self.doc.recompute()

        # Check if any paths originate inside inner hole of donut.  They should not.
        isInBox = False
        edges = [
            self.doc.Fusion.Shape.getElement(e) for e in ["Edge35", "Edge32", "Edge33", "Edge34"]
        ]
        square = Part.Wire(edges)
        sqrBB = square.BoundBox
        minPoint = FreeCAD.Vector(sqrBB.XMin, sqrBB.YMin, 0.0)
        maxPoint = FreeCAD.Vector(sqrBB.XMax, sqrBB.YMax, 0.0)
        for c in adaptive.Path.Commands:
            if pathOriginatesInBox(c, minPoint, maxPoint):
                isInBox = True
                break
        self.assertTrue(isInBox, "No paths originating within the inner hole.")

    def testModelStockAwareness(self):
        """testModelStockAwareness() Tests stock awareness- avoids cutting into the model regardless
        of bounding box selected."""
        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3", "Face10"])]  # (base, subs_list)
        adaptive.Label = "testModelStockAwareness+"
        adaptive.Comment = "testModelStockAwareness() Verify path generated on adjacent, combined Face3 and Face10.  The Z heights are different. Result should be the combination at Z=10 (faces from (0,0) to (40,25), minus tool radius), and only the lower face at Z=5: (15,0) to (40,25)."

        # Set additional operation properties
        setDepthsAndHeights(adaptive, 15, 0)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.ModelAwareExperiment = True
        adaptive.StepDown.Value = (
            5.0  # Have to set expression to None before numerical value assignment
        )
        # Don't use helix entry- ensures helix moves are counted in the path
        # boundary calculation. This should be unnecessary, as the helices are
        # grown out of the cut area, and thus must be inside of it.
        adaptive.UseHelixArcs = False

        _addViewProvider(adaptive)
        self.doc.recompute()

        # Check:
        # - Bounding box at Z=10 stays within Face3 and Face10- so -X for Face3,
        # +X and +/-Y for Face10
        # - bounding box at Z=5 stays within Face10
        # - No toolpaths at Z=0

        paths = [c for c in adaptive.Path.Commands if c.Name in ["G0", "G00", "G1", "G01"]]
        toolr = adaptive.OpToolDiameter.Value / 2
        tol = adaptive.Tolerance

        # Make clean up math below- combine tool radius and tolerance into a
        # single field that can be added/subtracted to/from bounding boxes
        moffset = toolr - tol

        zDict = getPathBoundaries(paths, [10, 5, 0])

        # NOTE: Face3 is at Z=10, Face10 is at Z=5
        bbf3 = self.doc.Fusion.Shape.getElement("Face3").BoundBox
        bbf10 = self.doc.Fusion.Shape.getElement("Face10").BoundBox

        okAt10 = (
            zDict[10] is not None
            and zDict[10]["min"][0] >= bbf3.XMin + moffset
            and zDict[10]["min"][1] >= bbf10.YMin + moffset
            and zDict[10]["max"][0] <= bbf10.XMax - moffset
            and zDict[10]["max"][1] <= bbf10.YMax - moffset
        )

        okAt5 = (
            zDict[5] is not None
            and zDict[5]["min"][0] >= bbf10.XMin + moffset
            and zDict[5]["min"][1] >= bbf10.YMin + moffset
            and zDict[5]["max"][0] < bbf10.XMax - moffset
            and zDict[5]["max"][1] < bbf10.YMax - moffset
        )

        okAt0 = not zDict[0]

        self.assertTrue(okAt10 and okAt5 and okAt0, "Path boundaries outside of expected regions")

    def testZStockToLeave(self):
        """testZStockToLeave() Tests Z stock to leave- with 1mm Z stock to leave, machining
        at the top of the model should not touch the top model face"""
        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, ["Face3", "Face10"])]  # (base, subs_list)
        adaptive.Label = "testZStockToLeave+"
        adaptive.Comment = "testZStockToLeave() Verify Z stock is left as requested"

        # Set additional operation properties
        setDepthsAndHeights(adaptive, 15, 10)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.StepDown.Value = (
            5.0  # Have to set expression to None before numerical value assignment
        )
        # Add some Z stock to leave so we avoid Face3 in this stepdown at Z=10
        adaptive.setExpression("ZStockToLeave", None)
        adaptive.ZStockToLeave.Value = 1

        _addViewProvider(adaptive)
        self.doc.recompute()

        # Check:
        # - No feed path at depth Z=10 touchs Face3
        toolr = adaptive.OpToolDiameter.Value / 2
        tol = adaptive.Tolerance

        # Make clean up math below- combine tool radius and tolerance into a
        # single field that can be added/subtracted to/from bounding boxes
        moffset = toolr - tol

        # Offset the face we don't expect to touch, verify no move is within
        # that boundary
        # NOTE: This isn't a perfect test (won't catch moves that start and end
        # outside of our face, but cut through/across it), but combined with
        # other tests should be sufficient.
        noPathTouchesFace3 = True
        foffset = self.doc.Fusion.Shape.getElement("Face3").makeOffset2D(moffset)
        # NOTE: Face3 is at Z=10, and the only feed moves will be at Z=10
        lastpt = FreeCAD.Vector(0, 0, 10)
        for p in [c.Parameters for c in adaptive.Path.Commands if c.Name in ["G1", "G01"]]:
            pt = FreeCAD.Vector(lastpt)
            if "X" in p:
                pt.x = p.get("X")
            if "Y" in p:
                pt.x = p.get("Y")

            if foffset.isInside(pt, 0.001, True):
                noPathTouchesFace3 = False
                break

            lastpt = pt

        self.assertTrue(noPathTouchesFace3, "No feed moves within the top face.")

    def testFullModelAdaptiveRoughing(self):
        """testFullModelAdaptiveRoughing() Tests full roughing- should machine entire model with no inputs"""
        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, [])]  # (base, subs_list)
        adaptive.Label = "testFullModelAdaptiveRoughing+"
        adaptive.Comment = (
            "testFullModelAdaptiveRoughing() Verify path generated with no subs roughs entire model"
        )

        # Set additional operation properties
        setDepthsAndHeights(adaptive, 15, 0)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.ModelAwareExperiment = True
        adaptive.StepDown.Value = (
            5.0  # Have to set expression to None before numerical value assignment
        )
        # Don't use helix entry- ensures helix moves are counted in the path
        # boundary calculation. This should be unnecessary, as the helices are
        # grown out of the cut area, and thus must be inside of it.
        adaptive.UseHelixArcs = False

        _addViewProvider(adaptive)
        self.doc.recompute()

        # Check:
        # - Bounding box at Z=0 goes outside the model box + tool diameter
        # (has to profile the model)
        # - Bounding box at Z=5 should go past the model in -X, but only up to the
        # stock edges in +X and Y
        # - Bounding box at Z=10 goes to at least stock bounding box edges,
        # minus tool diameter (has to machine the entire top of the stock off)
        # - [Should maybe check] At least one move Z = [10,5] is within the model
        # - [Should maybe check] No moves at Z = 0 are within the model

        paths = [c for c in adaptive.Path.Commands if c.Name in ["G0", "G00", "G1", "G01"]]
        toolr = adaptive.OpToolDiameter.Value / 2
        tol = adaptive.Tolerance

        # Make clean up math below- combine tool radius and tolerance into a
        # single field that can be added/subtracted to/from bounding boxes
        moffset = toolr - tol

        zDict = getPathBoundaries(paths, [10, 5, 0])
        mbb = self.doc.Fusion.Shape.BoundBox
        sbb = adaptive.Document.Stock.Shape.BoundBox

        okAt10 = (
            zDict[10] is not None
            and zDict[10]["min"][0] <= sbb.XMin + moffset
            and zDict[10]["min"][1] <= sbb.YMin + moffset
            and zDict[10]["max"][0] >= sbb.XMax - moffset
            and zDict[10]["max"][1] >= sbb.YMax - moffset
        )

        okAt5 = (
            zDict[5] is not None
            and zDict[5]["min"][0] <= mbb.XMin - moffset
            and zDict[5]["min"][1] <= sbb.YMin + moffset
            and zDict[5]["max"][0] >= sbb.XMax - moffset
            and zDict[5]["max"][1] >= sbb.YMax - moffset
        )

        okAt0 = (
            zDict[0] is not None
            and zDict[0]["min"][0] <= mbb.XMin - moffset
            and zDict[0]["min"][1] <= mbb.YMin - moffset
            and zDict[0]["max"][0] >= mbb.XMax + moffset
            and zDict[0]["max"][1] >= mbb.YMax + moffset
        )

        self.assertTrue(
            okAt10 and okAt5 and okAt0, "Path boundaries don't include expected regions"
        )

    def testStockLimitsAwareness(self):
        """testStockLimitsAwareness() Tests stock handling- should rough full model, but not cut
        air excessively where there's not stock"""
        # Instantiate a Adaptive operation and set Base Geometry
        adaptive = PathAdaptive.Create("Adaptive")
        adaptive.Base = [(self.doc.Fusion, [])]  # (base, subs_list)
        adaptive.Label = "testStockLimitsAwareness+"
        adaptive.Comment = (
            "testStockLimitsAwareness() Verify machining region is limited to the stock"
        )

        # Set additional operation properties
        setDepthsAndHeights(adaptive, 15, 5)
        adaptive.FinishingProfile = False
        adaptive.HelixAngle = 75.0
        adaptive.HelixDiameterLimit.Value = 1.0
        adaptive.LiftDistance.Value = 1.0
        adaptive.StepOver = 75
        adaptive.UseOutline = False
        adaptive.setExpression("StepDown", None)
        adaptive.ModelAwareExperiment = True
        adaptive.StepDown.Value = (
            5.0  # Have to set expression to None before numerical value assignment
        )
        # Don't use helix entry- ensures helix moves are counted in the path
        # boundary calculation. This should be unnecessary, as the helices are
        # grown out of the cut area, and thus must be inside of it.
        adaptive.UseHelixArcs = False

        # Create and assign new stock that will create different bounds at
        # different stepdowns
        btall = Part.makeBox(17, 27, 11, FreeCAD.Vector(-1, -1, 0))
        bshort = Part.makeBox(42, 27, 6, FreeCAD.Vector(-1, -1, 0))
        adaptive.Document.Job.Stock.Shape = btall.fuse(bshort)

        _addViewProvider(adaptive)
        # NOTE: Do NOT recompute entire doc, which will undo our stock change!
        adaptive.recompute()

        # Check:
        # - Bounding box at Z=10 stays basically above "btall"
        # - Bounding box at Z=5 and Z=0 are outside of stock

        paths = [c for c in adaptive.Path.Commands if c.Name in ["G1", "G01"]]
        toolr = adaptive.OpToolDiameter.Value / 2
        tol = adaptive.Tolerance

        # Make clean up math below- combine tool radius and tolerance into a
        # single field that can be added/subtracted to/from bounding boxes
        # NOTE: ADD tol here, since we're effectively flipping our normal
        # comparison and want tolerance to make our check looser
        moffset = toolr + tol

        zDict = getPathBoundaries(paths, [10, 5])
        sbb = adaptive.Document.Stock.Shape.BoundBox
        sbb10 = btall.BoundBox

        # These should be no more than a tool radius outside of the "btall"
        # XY section of the stock
        okAt10 = (
            zDict[10] is not None
            and zDict[10]["min"][0] >= sbb10.XMin - moffset
            and zDict[10]["min"][1] >= sbb10.YMin - moffset
            and zDict[10]["max"][0] <= sbb10.XMax + moffset
            and zDict[10]["max"][1] <= sbb10.YMax + moffset
        )

        # These should be no more than a tool radius outside of the overall
        # stock bounding box
        okAt5 = (
            zDict[5] is not None
            and zDict[5]["min"][0] >= sbb.XMin - moffset
            and zDict[5]["min"][1] >= sbb.YMin - moffset
            and zDict[5]["max"][0] <= sbb.XMax + moffset
            and zDict[5]["max"][1] <= sbb.YMax + moffset
        )

        self.assertTrue(okAt10 and okAt5, "Path feeds extend excessively in +X")

    # POSSIBLY MISSING TESTS:
    # - Something for region ordering
    # - Known-edge cases: cones/spheres/cylinders (especially partials on edges
    # of model + strange angles- especially for cylinders)
    # - Multiple models/stock
    # - XY stock to leave


# Eclass


def getPathBoundaries(paths, zLevels):
    """getPathBoundaries(paths, zLevels): Takes the list of paths and list of Z
    depths of interest, and finds the bounding box of the paths at each depth.
    A dictionary of depth: {"min": (x,y), "max": (x,y)} entries is returned.

    NOTE: You'd think that using Path.BoundBox would give us what we want,
    but... no, for whatever reason it appears to always extend to (0,0,0)
    """
    last = FreeCAD.Vector(0.0, 0.0, 0.0)
    # First make sure each element has X, Y, and Z coordinates
    for p in paths:
        params = p.Parameters
        last.x = p.X if "X" in params else last.x
        last.y = p.Y if "Y" in params else last.y
        last.z = p.Z if "Z" in params else last.z

        p.X = last.x
        p.Y = last.y
        p.Z = last.z

    zDict = {}
    for z in zLevels:
        zpaths = [k for k in paths if k.Z == z]
        if not zpaths:
            zDict[z] = None
            continue
        xmin = min([k.X for k in zpaths])
        xmax = max([k.X for k in zpaths])
        ymin = min([k.Y for k in zpaths])
        ymax = max([k.Y for k in zpaths])
        zDict[z] = {"min": (xmin, ymin), "max": (xmax, ymax)}

    return zDict


def setDepthsAndHeights(op, strDep=20.0, finDep=0.0):
    """setDepthsAndHeights(op, strDep=20.0, finDep=0.0)... Sets default depths and heights for `op` passed to it"""

    # Set start and final depth in order to eliminate effects of stock (and its default values)
    op.setExpression("StartDepth", None)
    op.StartDepth.Value = strDep
    op.setExpression("FinalDepth", None)
    op.FinalDepth.Value = finDep

    # Set step down so as to only produce one layer path
    op.setExpression("StepDown", None)
    op.StepDown.Value = 20.0

    # Set Heights
    # default values used


def getGcodeMoves(cmdList, includeRapids=True, includeLines=True, includeArcs=True):
    """getGcodeMoves(cmdList, includeRapids=True, includeLines=True, includeArcs=True)...
    Accepts command dict and returns point string coordinate.
    """

    # NOTE: Can NOT just check "if p.get("X")" or similar- that chokes when X is
    # zero. That becomes especially obvious when Z=0, and moves end up on the
    # wrong depth
    gcode_list = list()
    last = FreeCAD.Vector(0.0, 0.0, 0.0)
    for c in cmdList:
        p = c.Parameters
        name = c.Name
        if (includeRapids and name in ["G0", "G00"]) or (includeLines and name in ["G1", "G01"]):
            gcode = name
            x = last.x
            y = last.y
            z = last.z
            if "X" in p:
                x = round(p["X"], 2)
            gcode += " X" + str(x)
            if "Y" in p:
                y = round(p["Y"], 2)
            gcode += " Y" + str(y)
            if "Z" in p:
                z = round(p["Z"], 2)
            gcode += " Z" + str(z)
            last.x = x
            last.y = y
            last.z = z
            gcode_list.append(gcode)
        elif includeArcs and name in ["G2", "G3", "G02", "G03"]:
            gcode = name
            x = last.x
            y = last.y
            z = last.z
            i = 0.0
            j = 0.0
            k = 0.0
            if "I" in p:
                i = round(p["I"], 2)
            gcode += " I" + str(i)
            if "J" in p:
                j = round(p["J"], 2)
            gcode += " J" + str(j)
            if "K" in p:
                k = round(p["K"], 2)
            gcode += " K" + str(k)

            if "X" in p:
                x = round(p["X"], 2)
            gcode += " X" + str(x)
            if "Y" in p:
                y = round(p["Y"], 2)
            gcode += " Y" + str(y)
            if "Z" in p:
                z = round(p["Z"], 2)
            gcode += " Z" + str(z)

            gcode_list.append(gcode)
            last.x = x
            last.y = y
            last.z = z
    return gcode_list


def pathOriginatesInBox(cmd, minPoint, maxPoint):
    p = cmd.Parameters
    name = cmd.Name
    if name in ["G0", "G00", "G1", "G01"]:
        if "X" in p and "Y" in p:
            x = p.get("X")
            y = p.get("Y")
            if x > minPoint.x and y > minPoint.y and x < maxPoint.x and y < maxPoint.y:
                return True
    return False


def _addViewProvider(adaptiveOp):
    if FreeCAD.GuiUp:
        PathOpGui = PathAdaptiveGui.PathOpGui
        cmdRes = PathAdaptiveGui.Command.res
        adaptiveOp.ViewObject.Proxy = PathOpGui.ViewProvider(adaptiveOp.ViewObject, cmdRes)


# Example string literal of expected path moves from an operation
# Expected moves for unit test01
expected_moves_test01 = "G1 X32.5 Y32.5 Z5.0;  \
G1 X17.5 Y32.5 Z5.0;  \
G1 X17.5 Y30.0 Z5.0;  \
G1 X32.5 Y30.0 Z5.0;  \
G1 X32.5 Y27.5 Z5.0;  \
G1 X17.5 Y27.5 Z5.0;  \
G1 X17.5 Y25.0 Z5.0;  \
G1 X32.5 Y25.0 Z5.0;  \
G1 X32.5 Y22.5 Z5.0;  \
G1 X17.5 Y22.5 Z5.0;  \
G1 X17.5 Y20.0 Z5.0;  \
G1 X32.5 Y20.0 Z5.0;  \
G1 X32.5 Y17.5 Z5.0;  \
G1 X17.5 Y17.5 Z5.0"



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# Unit tests for the Arch wall module

import os
import Arch
import Draft
import Part
import FreeCAD as App
from bimtests import TestArchBase
from draftutils.messages import _msg


class TestArchRoof(TestArchBase.TestArchBase):

    def testRoof(self):
        operation = "Checking Arch Roof..."
        self.printTestMessage(operation)

        r = Draft.makeRectangle(length=2, height=-1)
        r.recompute()  # required before calling Arch.makeRoof
        ro = Arch.makeRoof(r)
        self.assertTrue(ro, "Arch Roof failed")

    def testRoof81Permutations(self):
        """Create 81 roofs using a range of arguments."""
        operation = "Arch Roof testRoof81Permutations"
        self.printTestMessage(operation)

        pts = [
            App.Vector(0, 0, 0),
            App.Vector(2000, 0, 0),
            App.Vector(4000, 0, 0),
            App.Vector(4000, 4000, 0),
            App.Vector(0, 4000, 0),
        ]
        ptsMod = [App.Vector(2000, 0, 0), App.Vector(2000, -1000, 0), App.Vector(2000, 1000, 0)]
        angsMod = [[60, 60], [30, 60], [60, 30]]
        runsMod = [[500, 500], [400, 500], [500, 400]]
        overhangsMod = [[100, 100], [100, 200], [200, 100]]
        delta = 6000
        pla = App.Placement()
        for iY in range(9):
            for iX in range(9):
                pts[1] = ptsMod[iY % 3]  # to get different edge angles
                angsLst = angsMod[iY // 3] + [90, 90, 90]
                runsLst = runsMod[iX % 3] + [0, 0, 0]
                overhangsLst = overhangsMod[iX // 3] + [0, 0, 0]
                pla.Base = App.Vector(iX * delta, iY * delta, 0)
                wire = Draft.makeWire(pts, closed=True)
                wire.MakeFace = False
                wire.Placement = pla
                wire.recompute()  # required before calling Arch.makeRoof
                roof = Arch.makeRoof(wire, angles=angsLst, run=runsLst, overhang=overhangsLst)
                roof.recompute()
                self.assertFalse(roof.Shape.isNull(), "'{}' failed".format(operation))
                self.assertTrue(roof.Shape.isValid(), "'{}' failed".format(operation))

    def testRoofAllAngles90(self):
        """Create a roof with the angles of all segments set at 90 degrees.
        This corner case results in a flat roof.
        """
        operation = "Arch Roof testRoofAllAngles90"
        self.printTestMessage(operation)

        pts = [
            App.Vector(0, 0, 0),
            App.Vector(2000, 0, 0),
            App.Vector(2000, 2000, 0),
            App.Vector(0, 2000, 0),
        ]

        wire = Draft.makeWire(pts, closed=True)
        wire.MakeFace = False
        wire.recompute()  # required before calling Arch.makeRoof
        roof = Arch.makeRoof(wire, angles=[90, 90, 90, 90])
        roof.recompute()
        self.assertFalse(roof.Shape.isNull(), "'{}' failed".format(operation))
        self.assertTrue(roof.Shape.isValid(), "'{}' failed".format(operation))

    def testRoofApex(self):
        """Create a hipped roof that relies on apex calculation. The roof has
        2 triangular segments with a single apex point.
        """
        operation = "Arch Roof testRoofApex"
        self.printTestMessage(operation)

        rec = Draft.makeRectangle(length=4000, height=3000, face=False)
        rec.recompute()  # required before calling Arch.makeRoof
        roof = Arch.makeRoof(
            rec,
            angles=[30, 40, 50, 60],
            run=[2000, 0, 2000, 0],
            idrel=[-1, 0, -1, 0],
            thickness=[50.0],
            overhang=[100.0],
        )
        roof.recompute()
        self.assertFalse(roof.Shape.isNull(), "'{}' failed".format(operation))
        self.assertTrue(roof.Shape.isValid(), "'{}' failed".format(operation))

    def testRoofSingleEavePoint(self):
        """Create a roof with a single triangular segment that has a single
        eave point.
        """
        operation = "Arch Roof testRoofSingleEavePoint"
        self.printTestMessage(operation)

        pts = [
            App.Vector(0, 0, 0),
            App.Vector(2000, 0, 0),
            App.Vector(4000, 2000, 0),
            App.Vector(-2000, 2000, 0),
        ]
        wire = Draft.makeWire(pts, closed=True)
        wire.MakeFace = False
        wire.recompute()  # required before calling Arch.makeRoof
        roof = Arch.makeRoof(
            wire, angles=[45, 90, 90, 90], run=[1000, 0, 0, 0], overhang=[1000, 0, 0, 0]
        )
        roof.recompute()
        self.assertFalse(roof.Shape.isNull(), "'{}' failed".format(operation))
        self.assertTrue(roof.Shape.isValid(), "'{}' failed".format(operation))

    def test_makeRoof(self):
        """Test the makeRoof function."""
        operation = "Testing makeRoof function"
        self.printTestMessage(operation)

        roof = Arch.makeRoof(name="TestRoof")
        self.assertIsNotNone(roof, "makeRoof failed to create a roof object.")
        self.assertEqual(roof.Label, "TestRoof", "Roof label is incorrect.")



# ***************************************************************************
# *   Copyright (c) 2016 Ofentse Kgoa <kgoaot@eskom.co.za>                  *
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Based on the FemElementGeometry1D by Bernd Hahnebach                  *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM _element fluid 1D document object"
__author__ = "Ofentse Kgoa"
__url__ = "https://www.freecad.org"

## @package element_fluid1D
#  \ingroup FEM
#  \brief element fluid 1D object

from . import base_femelement


class ElementFluid1D(base_femelement.BaseFemElement):
    """
    The element_fluid1D object
    """

    Type = "Fem::ElementFluid1D"

    known_fluid_types = ["Liquid"]
    # "Gas", "Open Channel" are not implemented in ccx writer
    # known_fluid_types = ["Liquid", "Gas", "Open Channel"]
    known_liquid_types = [
        "PIPE MANNING",
        "PIPE ENLARGEMENT",
        "PIPE CONTRACTION",
        "PIPE INLET",
        "PIPE OUTLET",
        "PIPE ENTRANCE",
        "PIPE DIAPHRAGM",
        "PIPE BEND",
        "PIPE GATE VALVE",
        "LIQUID PUMP",
        "PIPE WHITE-COLEBROOK",
    ]
    known_gas_types = ["NONE"]
    known_channel_types = ["NONE"]

    def __init__(self, obj):
        super().__init__(obj)

        obj.addProperty(
            "App::PropertyEnumeration", "SectionType", "FluidSection", "select fluid section type"
        )
        obj.setPropertyStatus("SectionType", "LockDynamic")

        obj.addProperty(
            "App::PropertyEnumeration",
            "LiquidSectionType",
            "LiquidSection",
            "select liquid section type",
        )
        obj.setPropertyStatus("LiquidSectionType", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "ManningArea",
            "LiquidManning",
            "set area of the manning fluid section",
        )
        obj.setPropertyStatus("ManningArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyLength",
            "ManningRadius",
            "LiquidManning",
            "set hydraulic radius of manning fluid section",
        )
        obj.setPropertyStatus("ManningRadius", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "ManningCoefficient",
            "LiquidManning",
            "set coefficient of manning fluid section",
        )
        obj.setPropertyStatus("ManningCoefficient", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "EnlargeArea1",
            "LiquidEnlargement",
            "set initial area of the enlargement fluid section",
        )
        obj.setPropertyStatus("EnlargeArea1", "LockDynamic")
        obj.addProperty(
            "App::PropertyArea",
            "EnlargeArea2",
            "LiquidEnlargement",
            "set enlarged area of enlargement fluid section",
        )
        obj.setPropertyStatus("EnlargeArea2", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "ContractArea1",
            "LiquidContraction",
            "set initial area of the contraction fluid section",
        )
        obj.setPropertyStatus("ContractArea1", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "ContractArea2",
            "LiquidContraction",
            "set contracted area of contraction fluid section",
        )
        obj.setPropertyStatus("ContractArea2", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "InletPressure",
            "LiquidInlet",
            "set inlet pressure for fluid section",
        )
        obj.setPropertyStatus("InletPressure", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "OutletPressure",
            "LiquidOutlet",
            "set outlet pressure for fluid section",
        )
        obj.setPropertyStatus("OutletPressure", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "InletFlowRate",
            "LiquidInlet",
            "set inlet mass flow rate for fluid section",
        )
        obj.setPropertyStatus("InletFlowRate", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "OutletFlowRate",
            "LiquidOutlet",
            "set outlet mass flow rate for fluid section",
        )
        obj.setPropertyStatus("OutletFlowRate", "LockDynamic")

        obj.addProperty(
            "App::PropertyBool",
            "InletPressureActive",
            "LiquidInlet",
            "activates or deactivates inlet pressure for fluid section",
        )
        obj.setPropertyStatus("InletPressureActive", "LockDynamic")

        obj.addProperty(
            "App::PropertyBool",
            "OutletPressureActive",
            "LiquidOutlet",
            "activates or deactivates outlet pressure for fluid section",
        )
        obj.setPropertyStatus("OutletPressureActive", "LockDynamic")

        obj.addProperty(
            "App::PropertyBool",
            "InletFlowRateActive",
            "LiquidInlet",
            "activates or deactivates inlet flow rate for fluid section",
        )
        obj.setPropertyStatus("InletFlowRateActive", "LockDynamic")

        obj.addProperty(
            "App::PropertyBool",
            "OutletFlowRateActive",
            "LiquidOutlet",
            "activates or deactivates outlet flow rate for fluid section",
        )
        obj.setPropertyStatus("OutletFlowRateActive", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "EntrancePipeArea",
            "LiquidEntrance",
            "set the pipe area of the entrance fluid section",
        )
        obj.setPropertyStatus("EntrancePipeArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "EntranceArea",
            "LiquidEntrance",
            "set the entrance area of the entrance fluid section",
        )
        obj.setPropertyStatus("EntranceArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "DiaphragmPipeArea",
            "LiquidDiaphragm",
            "set the pipe area of the diaphragm fluid section",
        )
        obj.setPropertyStatus("DiaphragmPipeArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "DiaphragmArea",
            "LiquidDiaphragm",
            "set the diaphragm area of the diaphragm fluid section",
        )
        obj.setPropertyStatus("DiaphragmArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "BendPipeArea",
            "LiquidBend",
            "set pipe area of the bend fluid section",
        )
        obj.setPropertyStatus("BendPipeArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "BendRadiusDiameter",
            "LiquidBend",
            "set ratio of bend radius over pipe diameter of the bend fluid section",
        )
        obj.setPropertyStatus("BendRadiusDiameter", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "BendAngle",
            "LiquidBend",
            "set bend angle of the bend fluid section",
        )
        obj.setPropertyStatus("BendAngle", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "BendLossCoefficient",
            "LiquidBend",
            "set loss coefficient of the bend fluid section",
        )
        obj.setPropertyStatus("BendLossCoefficient", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "GateValvePipeArea",
            "LiquidGateValve",
            "set pipe area of the gate valve fluid section",
        )
        obj.setPropertyStatus("GateValvePipeArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "GateValveClosingCoeff",
            "LiquidGateValve",
            "set closing coefficient of the gate valve fluid section",
        )
        obj.setPropertyStatus("GateValveClosingCoeff", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloatList",
            "PumpFlowRate",
            "LiquidPump",
            "set the pump characteristic flow rate of the pump fluid section",
        )
        obj.setPropertyStatus("PumpFlowRate", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloatList",
            "PumpHeadLoss",
            "LiquidPump",
            "set the pump characteristic head loss of the pump fluid section",
        )
        obj.setPropertyStatus("PumpHeadLoss", "LockDynamic")

        obj.addProperty(
            "App::PropertyArea",
            "ColebrookeArea",
            "LiquidColebrooke",
            "set pipe area of the colebrooke fluid section",
        )
        obj.setPropertyStatus("ColebrookeArea", "LockDynamic")

        obj.addProperty(
            "App::PropertyLength",
            "ColebrookeRadius",
            "LiquidColebrooke",
            "set hydraulic radius of the colebrooke fluid section",
        )
        obj.setPropertyStatus("ColebrookeRadius", "LockDynamic")

        obj.addProperty(
            "App::PropertyLength",
            "ColebrookeGrainDiameter",
            "LiquidColebrooke",
            "set grain diameter of the colebrooke fluid section",
        )
        obj.setPropertyStatus("ColebrookeGrainDiameter", "LockDynamic")

        obj.addProperty(
            "App::PropertyFloat",
            "ColebrookeFormFactor",
            "LiquidColebrooke",
            "set coefficient of the colebrooke fluid section",
        )
        obj.setPropertyStatus("ColebrookeFormFactor", "LockDynamic")

        obj.addProperty(
            "App::PropertyEnumeration", "GasSectionType", "GasSection", "select gas section type"
        )
        obj.setPropertyStatus("GasSectionType", "LockDynamic")

        obj.addProperty(
            "App::PropertyEnumeration",
            "ChannelSectionType",
            "ChannelSection",
            "select channel section type",
        )
        obj.setPropertyStatus("ChannelSectionType", "LockDynamic")

        # set property default values
        obj.SectionType = ElementFluid1D.known_fluid_types
        obj.SectionType = "Liquid"
        obj.LiquidSectionType = ElementFluid1D.known_liquid_types
        obj.LiquidSectionType = "PIPE INLET"
        obj.GasSectionType = ElementFluid1D.known_gas_types
        obj.GasSectionType = "NONE"
        obj.ChannelSectionType = ElementFluid1D.known_channel_types
        obj.ChannelSectionType = "NONE"
        obj.ManningArea = 10.0
        obj.ManningRadius = 1.0
        obj.ManningCoefficient = 0.0015  # has units of s/mm^(1/3)
        obj.EnlargeArea1 = 10.0
        obj.EnlargeArea2 = 20.0
        obj.ContractArea1 = 20.0
        obj.ContractArea2 = 10.0
        obj.EntrancePipeArea = 20.0
        obj.EntranceArea = 20.0
        obj.DiaphragmPipeArea = 20.0
        obj.DiaphragmArea = 20.0
        obj.BendPipeArea = 20.0
        obj.BendRadiusDiameter = 1.0
        obj.BendAngle = 0.0
        obj.BendLossCoefficient = 0.0
        obj.GateValvePipeArea = 20.0
        obj.GateValveClosingCoeff = 0.125
        obj.PumpFlowRate = [0, 1.04e-04, 2.08e-4, 3.13e-4, 4.17e-4]
        obj.PumpHeadLoss = [30, 29.17, 26.67, 23.33, 18.33]
        obj.ColebrookeArea = 20.0
        obj.ColebrookeRadius = 1.0
        obj.ColebrookeGrainDiameter = 0.0025
        obj.ColebrookeFormFactor = 1.0
        obj.InletPressure = 1.0
        obj.OutletPressure = 1.0
        obj.InletFlowRate = 1.0
        obj.OutletFlowRate = 1.0
        obj.InletPressureActive = True
        obj.OutletPressureActive = True
        obj.InletFlowRateActive = False
        obj.OutletFlowRateActive = False
